# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Miscellaneous commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
from os.path import exists, isdir
from shutil import rmtree
from typing import List

# 3rd party imports
import click
import cliform

# local imports
from peltak.core import conf
from peltak.core import context
from peltak.core import fs
from peltak.core import log
from peltak.core import shell
from peltak.core import util
from peltak.core import templates


def clean(exclude):
    # type: (List[str]) -> None
    """ Remove all unnecessary files.

    Args:
        exclude (list[str]):
            A list of path patterns to exclude from deletion.
    """
    pretend = context.get('pretend', False)
    exclude = list(exclude) + conf.get('clean.exclude', [])
    clean_patterns = conf.get('clean.patterns', [
        '*__pycache__*',
        '*.py[cod]',
        '*.swp',
    ])

    num_files = 0
    with util.timed_block() as t:
        files = fs.filtered_walk(conf.proj_path(), clean_patterns, exclude)
        for path in files:
            try:
                num_files += 1

                if not isdir(path):
                    log.info('  <91>[file] <90>{}', path)
                    if not pretend:
                        os.remove(path)
                else:
                    log.info('  <91>[dir]  <90>{}', path)
                    if not pretend:
                        rmtree(path)

            except OSError:
                log.info("<33>Failed to remove <90>{}", path)

    if pretend:
        msg = "Would delete <33>{}<32> files. Took <33>{}<32>s"
    else:
        msg = "Deleted <33>{}<32> files in <33>{}<32>s"

    log.info(msg.format(num_files, t.elapsed_s))


class InitForm(cliform.Form):
    """ Everything needed to generate initial pelconf.yaml. """
    src_dir = cliform.Field(
        'Source directory',
        default='src',
        type=str,
        help=('The root directory for all your sources. This is what you '
              'would treat as PYTHONPATH')
    )
    version_file = cliform.Field(
        'Version file',
        type=str,
        default='',
        help=('This will be used for a lot of default values in the config '
              '(like for linting, reference docs etc.)')
    )


def init(quick, blank, force):
    # type: (bool, bool, bool) -> None
    """ Create an empty pelconf.yaml from template """
    config_file = 'pelconf.yaml'
    prompt = "-- <35>{} <32>already exists. Wipe it?<0>".format(config_file)

    if not force and exists(config_file) and not click.confirm(shell.fmt(prompt)):
        log.info("Canceled")
        return

    ctx = dict(blank=blank)

    if not blank:
        form = InitForm().run(quick=quick)
        ctx.update(form.values)

    confg_content = templates.Engine().render_file('pelconf.yaml', ctx)

    log.info('Writing <35>{}'.format(config_file))
    fs.write_file(config_file, confg_content)

    if context.get('verbose') > 0:
        for line in shell.highlight(confg_content, 'yaml').splitlines():
            log.info('  {}', line)


# Used in docstrings only until we drop python2 support
del List



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.versioning
    :synopsis: Functionality related to project versioning.
"""
from __future__ import absolute_import

# stdlib imports
import json
import re
from collections import OrderedDict
from os.path import exists
from typing import Optional, Tuple

from . import conf
from . import fs


RE_PY_VERSION = re.compile(
    r'__version__\s*=\s*["\']'
    r'(?P<version>\d+(\.\d+(\.\d+)?)?)'
    r'["\']'
)


# MAJOR.MINOR[.PATCH[-BUILD]]
RE_VERSION = re.compile(
    r'^'
    r'(?P<major>\d+)'
    r'(\.(?P<minor>\d+)'
    r'(\.(?P<patch>\d+))?)?'
    r'$'
)


def is_valid(version_str):
    # type: (str) -> bool
    """ Check if the given string is a version string

    Args:
        version_str (str):
            A string to check

    Returns:
        bool: **True** if the given string is a version.
    """
    return bool(version_str and RE_VERSION.match(version_str))


def current():
    # type: () -> str
    """ Return the project's current version.

    Returns:
        str: The current project version in MAJOR.MINOR.PATCH format. PATCH
        might be omitted if it's 0, so 1.0.0 becomes 1.0 and 0.1.0 becomes 0.1.
    """
    storage = get_version_storage()
    return storage.read() or ''


def write(version):
    # type: (str) -> None
    """ Write the given version to the VERSION_FILE """
    if not is_valid(version):
        raise ValueError("Invalid version: '{}'".format(version))

    storage = get_version_storage()
    storage.write(version)


def bump(component='patch', exact=None):
    # type: (str, str) -> Tuple[str, str]
    """ Bump the given version component.

    Args:
        component (str):
            What part of the version should be bumped. Can be one of:

            - major
            - minor
            - patch

        exact (str):
            The exact version that should be set instead of bumping the current
            one.

    Returns:
        tuple(str, str): A tuple of old and bumped version.
    """
    old_ver = current()

    if exact is None:
        new_ver = _bump_version(old_ver, component)
    else:
        new_ver = exact

    write(new_ver)
    return old_ver, new_ver


def _bump_version(version, component='patch'):
    # type: (str, str) -> str
    """ Bump the given version component.

    Args:
        version (str):
            The current version. The format is: MAJOR.MINOR[.PATCH].
        component (str):
            What part of the version should be bumped. Can be one of:

            - major
            - minor
            - patch

    Returns:
        str: Bumped version as a string.
    """
    if component not in ('major', 'minor', 'patch'):
        raise ValueError("Invalid version component: {}".format(component))

    m = RE_VERSION.match(version)
    if m is None:
        raise ValueError("Version must be in MAJOR.MINOR[.PATCH] format")

    major = m.group('major')
    minor = m.group('minor') or '0'
    patch = m.group('patch') or None

    if component == 'major':
        major = str(int(major) + 1)
        minor = '0'
        patch = '0'

    elif component == 'minor':
        minor = str(int(minor) + 1)
        patch = '0'

    else:
        patch = patch or '0'
        patch = str(int(patch) + 1)

    new_ver = '{}.{}.{}'.format(major, minor, patch)

    return new_ver


class VersionStorage(object):
    """ Base class for version storage.

    A version storage is a way to store the project version. Different projects
    will have different ways of storing the version. The simplest case is a
    text file that holds just the version number (probably called VERSION).
    Python projects can also use a version stored as ``__version__`` variable
    inside the project package/module. A Node.js project on the other hand will
    probably keep the version in ``package.json``. All of the above strategies
    can be (and are) implemented through subclassing this class.

    @see `PyVersionStorage`, `RawVersionStorage`, `NodeVersionStorage`
    """
    def __init__(self, version_file):
        # type: (str) -> None
        self.version_file = version_file

        if not exists(version_file):
            raise ValueError("Version file '{}' does not exist.".format(
                version_file
            ))

    def read(self):
        # type: () -> Optional[str]
        """ Read the current project version.

        All subclasses must implement this method.

        Raises:
            NotImplementedError:
                If the subclass does nto implement the read() method.
        """
        raise NotImplementedError("{} must implement .read()".format(
            self.__class__.__name__
        ))

    def write(self, version):
        # type: (str) -> None
        """ Save the given version as the current project version.

        All subclasses must implement this method.

        Raises:
            NotImplementedError:
                If the subclass does nto implement the write() method.
        """
        raise NotImplementedError("{} must implement .write()".format(
            self.__class__.__name__
        ))


class PyVersionStorage(VersionStorage):
    """ Store project version in one of the py module/package files. """
    def read(self):
        # type: () -> Optional[str]
        """ Read the project version from .py file.

        This will regex search in the file for a
        ``__version__ = VERSION_STRING`` and read the version string.
        """
        with open(self.version_file) as fp:
            content = fp.read()
            m = RE_PY_VERSION.search(content)
            if not m:
                return None
            else:
                return m.group('version')

    def write(self, version):
        # type: (str) -> None
        """ Write the project version to .py file.

        This will regex search in the file for a
        ``__version__ = VERSION_STRING`` and substitute the version string
        for the new version.
        """
        with open(self.version_file) as fp:
            content = fp.read()

        ver_statement = "__version__ = '{}'".format(version)
        new_content = RE_PY_VERSION.sub(ver_statement, content)
        fs.write_file(self.version_file, new_content)


class RawVersionStorage(VersionStorage):
    """ Store project version as a simple value in a text file. """
    def read(self):
        # type: () -> Optional[str]
        """ Read the project version from .py file.

        This will regex search in the file for a
        ``__version__ = VERSION_STRING`` and read the version string.
        """
        with open(self.version_file) as fp:
            version = fp.read().strip()

            if is_valid(version):
                return version

            return None

    def write(self, version):
        # type: (str) -> None
        fs.write_file(self.version_file, version)


class NodeVersionStorage(VersionStorage):
    """ Store project version in package.json. """
    def read(self):
        # type: () -> Optional[str]
        with open(self.version_file) as fp:
            config = json.load(fp)
            return config.get('version')

    def write(self, version):
        # type: (str) -> None
        with open(self.version_file, 'r') as fp:
            config = json.load(fp, object_pairs_hook=OrderedDict)

        config['version'] = version

        fs.write_file(self.version_file, json.dumps(config, indent=2) + '\n')


def get_version_storage():
    # type: () -> VersionStorage
    """ Get version storage for the given version file.

    The storage engine used depends on the extension of the *version.file* conf
    variable.
    """
    version_file = conf.get_path('version.file', 'VERSION')
    if version_file.endswith('.py'):
        return PyVersionStorage(version_file)
    elif version_file.endswith('package.json'):
        return NodeVersionStorage(version_file)
    else:
        return RawVersionStorage(version_file)


# Used in docstrings only until we drop python2 support
del Optional, Tuple



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.log
    :synopsis: Helpers for nice shell output formatting.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any

# local imports
from . import shell


def info(msg, *args, **kw):
    # type: (str, *Any, **Any) -> None
    """ Print sys message to stdout.

    System messages should inform about the flow of the script. This should
    be a major milestones during the build.
    """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    shell.cprint('-- <32>{}<0>'.format(msg))


def err(msg, *args, **kw):
    # type: (str, *Any, **Any) -> None
    """ Per step status messages

    Use this locally in a command definition to highlight more important
    information.
    """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    shell.cprint('-- <31>{}<0>'.format(msg))


# Used in docstrings only until we drop python2 support
del Any



# -*- coding: utf-8 -*-
""" Runtime context implementation.

.. module:: peltak.core.context
    :synopsis: Runtime context for peltak commands.


Runtime context is a unified place to store dynamic global configuration. The
verbosity and ``pretend`` is a good example of that.

Works like `peltak.core.pelconf` but the configuration can be dynamically modified
in runtime.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any, Text

# local imports
from . import util


class ContextError(Exception):
    """ Base class for context related exceptions. """
    pass


class InvalidPath(ContextError):
    """ Invalid variable path. """
    def __init__(self, path):
        super(InvalidPath, self).__init__("{} is not a dict".format(
            '.'.join(path)
        ))


class RunContext(util.Singleton):
    """ Runtime context.

    This class is the equivalent of conf but for values that can be modified
    in runtime. This is for all the settings that can be set on the command
    line and can span many commands or APIs.
    """

    def __init__(self):
        # __init__ can be called multiple times since this class is a singleton
        if not self._singleton_initialized:
            self.values = {}

    def clear(self):
        # type: () -> None
        """ Clear all existing values. """
        self.values = {}

    def get(self, name, *default):
        # type: (str, Any) -> Any
        """ Get context value with the given name and optional default.

        Args:
            name (str):
                The name of the context value.
            *default (Any):
                If given and the key doesn't not exist, this will be returned
                instead. If it's not given and the context value does not exist,
                `AttributeError` will be raised

        Returns:
            The requested context value.  If the value does not exist it will
            return *default* if give or raise `AttributeError`.

        Raises:
            AttributeError: If the value does not exist and *default* was not
                given.
        """

        curr = self.values
        for part in name.split('.'):
            if part in curr:
                curr = curr[part]
            elif default:
                return default[0]
            else:
                fmt = "Context value '{}' does not exist:\n{}"
                raise AttributeError(fmt.format(
                    name, util.yaml_dump(self.values)
                ))

        return curr

    def set(self, name, value):
        """ Set context value.

        Args:
            name (str):
                The name of the context value to change.
            value (Any):
                The new value for the selected context value
        """
        curr = self.values
        parts = name.split('.')

        for i, part in enumerate(parts[:-1]):
            try:
                curr = curr.setdefault(part, {})
            except AttributeError:
                raise InvalidPath('.'.join(parts[:i + 1]))

        try:
            curr[parts[-1]] = value
        except TypeError:
            raise InvalidPath('.'.join(parts[:-1]))


def get(name, *default):
    # type: (Text, *Any) -> Any
    """ Get context value with the given name and optional default.

    Args:
        name (str):
            The name of the context value.
        *default (Any):
            If given and the key doesn't not exist, this will be returned
            instead. If it's not given and the context value does not exist,
            `AttributeError` will be raised

    Returns:
        The requested context value.  If the value does not exist it will
        return *default* if give or raise `AttributeError`.

    Raises:
        AttributeError: If the value does not exist and *default* was not
            given.
    """
    return RunContext().get(name, *default)


def set(name, value):   # pylint: disable=redefined-builtin
    # type: (Text, Any) -> None
    """ Set context value.

    Args:
        name (str):
            The name of the context value to change.
        value (Any):
            The new value for the selected context value
    """
    RunContext().set(name, value)


def clear():
    # type: () -> None
    """ Clear all existing values. """
    return RunContext().clear()


# Used in type hints comments only (until we drop python2 support)
del Any, Text



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Types and classes used by ``peltak.extra.scripts``. """

# stdlib imports
from typing import Any, Callable, Dict, List, Optional, Type, cast

# 3rd party imports
import attr

# local imports
from six import string_types
from peltak.commands import click, pretend_option, verbose_option
from peltak.core import types


AnyFn = Callable[..., Any]
YamlConf = Dict[str, Any]
CliOptions = Dict[str, Any]


@attr.s
class ScriptOption(object):
    """ Describes the scripts command line option. """
    name = attr.ib(type=List[str])
    default = attr.ib(type=Any, default=None)
    about = attr.ib(type=str, default='')
    is_flag = attr.ib(type=bool, default=False)
    count = attr.ib(type=bool, default=False)
    type = attr.ib(type=Type, default=str)

    @classmethod
    def from_config(cls, option_conf):
        # type: (YamlConf) -> ScriptOption
        """ Load script option from configuration in pelconf.yaml """
        fields = attr.fields(cls)
        name = option_conf.get('name')

        if not name:
            raise ValueError("You must define the name of the option")

        if isinstance(name, string_types):
            # Support passing name: ['--opt'] or name: '--opt'
            name = [name]

        # Convert string type to an initializer for that type.
        if 'type' not in option_conf:
            opt_type = fields.type.default
        else:
            opt_type = {
                'str': str,
                'int': int,
                'float': float,
            }.get(option_conf['type'])

            if not opt_type:
                raise ValueError("Unsupported {} option type {}".format(
                    name, option_conf['type']
                ))

        return cls(
            name=name,
            default=option_conf.get('default', fields.default.default),
            about=option_conf.get('about', fields.about.default),
            is_flag=option_conf.get('is_flag', fields.is_flag.default),
            count=option_conf.get('count', fields.count.default),   # type: ignore
            type=cast(Type, opt_type),
        )


@attr.s
class Script(object):
    """ Represents a single script defined in pelconf.yaml.

    The important thing is to do a little as possible during creation and only
    do heavy stuff inside the `Script.run()` method. This is because the
    scripts are parsed in a *post-conf-load* hook and that code impacts the
    speed of shell auto completion for peltak command.
    """
    name = attr.ib(type=str)
    command = attr.ib(type=str)
    about = attr.ib(type=str, default='')
    root_cli = attr.ib(type=bool, default=False)
    success_exit_codes = attr.ib(type=List[int], factory=lambda: [0])
    options = attr.ib(type=List[ScriptOption], factory=list)
    files = attr.ib(type=Optional[types.FilesCollection], default=None)

    @classmethod
    def from_config(cls, name, script_conf):
        # type: (str, YamlConf) -> Script
        """ Load script from pelfconf.yaml """
        fields = attr.fields(cls)
        options = script_conf.get('options',
                                  fields.options.default.factory())  # type: ignore
        files = None
        success_exit_codes = script_conf.get(
            'success_exit_codes',
            fields.success_exit_codes.default.factory()   # type: ignore
        )

        # Parse 'files' section if it's present
        if 'files' in script_conf:
            files = types.FilesCollection.from_config(script_conf['files'])

        if isinstance(success_exit_codes, int):
            success_exit_codes = [success_exit_codes]

        # Cannot have a script without a 'command'.
        if 'command' not in script_conf:
            raise ValueError("Missing 'command' for '{}' script".format(name))

        return cls(
            name=name,
            command=script_conf['command'],
            about=script_conf.get('about', fields.about.default),
            root_cli=script_conf.get('root_cli', fields.root_cli.default),
            success_exit_codes=success_exit_codes,
            options=[ScriptOption.from_config(opt_conf) for opt_conf in options],
            files=files,
        )

    def register(self, cli_group):
        # type: (click.Group) -> None
        """ Register the script with click. """
        @verbose_option
        @pretend_option
        @click.pass_context
        def script_command(ctx, **options):  # pylint: disable=missing-docstring
            from .logic import run_script
            run_script(self, options)

        script_command.__doc__ = self.about

        # Add all option definitions to the generated click command.
        for option in self.options:
            script_command = self._add_option(script_command, option)

        cli_group.command(self.name)(script_command)

    def _add_option(self, cmd_fn, option):
        # type: (AnyFn, ScriptOption) -> AnyFn
        return click.option(
            *option.name,
            is_flag=option.is_flag,
            default=option.default,
            help=option.about,
            count=option.count,
            type=option.type
        )(cmd_fn)



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Django related management commands.

Those commands replace the usage of ``./manage.py`` (thus it's removed). Those
correspond 1 to 1 to their ``./manage.py`` counterparts. Those commands mainly
exists so that manage.py can be deleted (less top-level files in project dir).
"""
from __future__ import absolute_import

from peltak.commands import root_cli, click, pretend_option
from peltak.core import conf


conf.command_requirements(
    'django~=1.11'
)


@root_cli.group('django')
def django_cli():
    # type: () -> None
    """ Commands related to django """
    pass


@django_cli.command('devserver')
@click.option(
    '-p', '--port',
    type=int,
    default=8000,
    help="Port the server will run on"
)
@click.option(
    '-s', '--settings',
    type=str,
    help="Settings module to use. Defaults to DJANGO_SETTINGS conf variable."
)
@pretend_option
def devserver(port=8000, settings=None):
    # type: (int, str) -> None
    """ Run dev server. """
    _manage_cmd(['runserver', '0.0.0.0:{}'.format(port)], settings)


@django_cli.command('collectstatic')
@pretend_option
def collectstatic():
    # type: () -> None
    """ Collect all static files.

    This command won't ask for permission as manage.py does.

    Sample Config::

        \b
        src_dir: 'src'
        django:
          settings: 'mypkg.settings'

    Example::

        $ peltak django collectstatic

    """
    _manage_cmd(['collectstatic', '--no-input'])


@django_cli.command('mkmigrations')
@click.argument('app')
@click.argument('name')
@pretend_option
def mkmigrations(app, name):
    # type: (str, str) -> None
    """ Create a named migration for a given app.

    This will require the user to name every migration he creates thus improving
    the source code. Otherwise it's just a simple wrapper around ./manage.py.

    Sample Config::

        \b
        src_dir: 'src'
        django:
          settings: 'mypkg.settings'

    Example::

        $ peltak django mkmigrations mypkg.app add_new_fields_to_mytable

    """
    _manage_cmd(['makemigrations', '-n', name] + app.split(' '))


@django_cli.command('migrate')
@pretend_option
def migrate():
    # type: () -> None
    """ Apply pending migrations.

    Sample Config::

        \b
        src_dir: 'src'
        django:
          settings: 'mypkg.settings'

    Example::

        $ peltak django migrate

    """
    _manage_cmd(['migrate'])


@django_cli.command('createsuperuser')
@pretend_option
def createsuperuser():
    """ Create super user (probably needed for the first user).

    Sample Config::

        \b
        src_dir: 'src'

        django:
          settings: 'mypkg.settings'

    Example::

        $ peltak django createsuperuser

    """
    _manage_cmd(['createsuperuser'])


@django_cli.command('shell')
@pretend_option
def shell():
    # type: () -> None
    """ Start django shell

    Sample Config::

        \b
        src_dir: 'src'
        django:
          settings: 'mypkg.settings'

    Example::

        $ peltak django shell

    """
    _manage_cmd(['shell'])


def _manage_cmd(cmd, settings=None):
    # type: () -> None
    """ Run django ./manage.py command manually.

    This function eliminates the need for having ``manage.py`` (reduces file
    clutter).
    """
    import sys
    from os import environ
    from peltak.core import conf
    from peltak.core import context
    from peltak.core import log

    sys.path.insert(0, conf.get('src_dir'))

    settings = settings or conf.get('django.settings', None)
    environ.setdefault("DJANGO_SETTINGS_MODULE", settings)

    args = sys.argv[0:-1] + cmd

    if context.get('pretend', False):
        log.info("Would run the following manage command:\n<90>{}", args)
    else:
        from django.core.management import execute_from_command_line
        execute_from_command_line(args)



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
##################
``peltak feature``
##################

git flow feature commands.
"""
from peltak.commands import root_cli, click, pretend_option


@root_cli.group('feature', invoke_without_command=True)
def feature_cli():
    # type: () -> None
    """ Start a new git-flow feature.  """


@feature_cli.command('start')
@click.argument('name', required=False)
@pretend_option
def start(name):
    # type: (str) -> None
    """ Start a new git-flow feature.  """
    from peltak.extra.gitflow import logic

    if name is None:
        name = click.prompt('Feature name')

    logic.feature.start(name)


@feature_cli.command('rename')
@click.argument('name', required=False)
@pretend_option
def rename(name):
    # type: (str) -> None
    """ Give the currently developed feature a new name. """
    from peltak.extra.gitflow import logic

    if name is None:
        name = click.prompt('Feature name')

    logic.feature.rename(name)


@feature_cli.command('update')
@pretend_option
def update():
    # type: () -> None
    """ Update the feature with updates committed to develop. """
    from peltak.extra.gitflow import logic
    logic.feature.update()


@feature_cli.command('finish')
@pretend_option
@click.option(
    '--ff', '--fast-forward', 'fast_forward',
    is_flag=True,
    help="Try to perform a fast-forward merge. If possible this will not "
         "create a merge commit on the target branch."
)
def finish(fast_forward):
    # type: (bool) -> None
    """ Merge current feature into develop. """
    from peltak.extra.gitflow import logic
    logic.feature.finish(fast_forward)


@feature_cli.command('merged')
@pretend_option
def merged():
    # type: () -> None
    """ Cleanup a remotely merged branch. """
    from peltak.extra.gitflow import logic
    logic.feature.merged()



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" pypi related commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
from os.path import join

# local imports
from peltak.core import conf
from peltak.core import fs
from peltak.core import log
from peltak.core import shell
from peltak.core import util


def upload(target):
    # type: (str) -> None
    """ Upload the release to a pypi server.

    TODO: Make sure the git directory is clean before allowing a release.

    Args:
        target (str):
            pypi target as defined in ~/.pypirc
    """
    log.info("Uploading to pypi server <33>{}".format(target))
    with conf.within_proj_dir():
        shell.run('python setup.py sdist register -r "{}"'.format(target))
        shell.run('python setup.py sdist upload -r "{}"'.format(target))


def gen_pypirc(username=None, password=None):
    # type: (str, str) -> None
    """ Generate ~/.pypirc with the given credentials.

    Useful for CI builds. Can also get credentials through env variables
    ``PYPI_USER`` and ``PYPI_PASS``.

    Args:
        username (str):
            pypi username. If not given it will try to take it from the
            `` PYPI_USER`` env variable.
        password (str):
            pypi password. If not given it will try to take it from the
            `` PYPI_PASS`` env variable.
    """
    path = join(conf.get_env('HOME'), '.pypirc')
    username = username or conf.get_env('PYPI_USER', None)
    password = password or conf.get_env('PYPI_PASS', None)

    if username is None or password is None:
        log.err("You must provide $PYPI_USER and $PYPI_PASS")
        sys.exit(1)

    log.info("Generating <94>{}".format(path))

    fs.write_file(path, util.remove_indent('''
        [distutils]
        index-servers = pypi
        
        [pypi]
        repository: https://upload.pypi.org/legacy/
        username: {username}
        password: {password}
        
    '''.format(
        username=username,
        password=password
    )))



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Simple command line form/wizard implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any, Callable, Dict, Tuple

# 3rd party imports
import click
from six import with_metaclass

# local imports
from peltak.core import shell


class Field(object):
    """ Defines a field on a form.

    Each field has a unique ID that can be used to identify it on the form.

    Attributes:
        id (str):
            Field ID. The ID will be used in the form as the key for the user
            input. This should be unique across the form.
        prompt (str):
            The text of the prompt displayed to the user. Should not include
            any styling like colors as this will be handled automatically.
        type (Callable):
            The user input type converter. This should be a callable that when
            called with the string provided by the user should return the python
            representation of the value or raise `ValueError` if user input is
            invalid.
        default (Any):
            The default value. If the form is ran in quick mode, it will be used
            instead of asking the user.
        help (str):
            Help text displayed above the prompt. This should describe what the
            value is used for so the user can better understand what's required
            of him.
    """
    _idx = 0

    def __init__(self, prompt, type, help, default=None):
        # type: (str, Callable, str, Any) -> None
        self._idx += 1

        self.id = None
        self.form = None

        self.prompt = prompt
        self.type = type
        self.help = help
        self.default = default

    def bind(self, field_id):
        """ Bind this field to the given ID.

        The field ID is usually the name of the field attribute as defined in
        the final form so cannot be provided in the constructor. The ID will
        be extracted from attribute name by `FormMeta` metaclass.
        """
        self.id = field_id
        return self

    @property
    def pretty_prompt(self):
        # type: () -> str
        """ Return a colorized prompt ready to be displayed to the user. """
        return shell.fmt('<1>{}<0>'.format(self.prompt))


class FormMeta(type):
    """ Form metaclass.

    It will bind all the fields in the form and make them available as
    ``cls.fields`` attribute for easy access.
    """
    def __init__(cls, name, bases, attrs):
        # type: (str, Tuple[type], Dict[str, Any]) -> None
        super(FormMeta, cls).__init__(name, bases, attrs)

        cls.fields = sorted(
            (field.bind(name) for name, field in attrs.items()
             if isinstance(field, Field)),
            key=lambda x: x._idx        # noqa
        )


class Form(with_metaclass(FormMeta)):
    """ CLI form.

    This class makes it easy to create a CLI forms/wizards when you need to
    ask the user for more details in order to continue.
    """
    def __init__(self):
        # type: () -> None
        self.values = {}

    def run(self, quick=False):
        # type: (bool) -> Form
        for field in self.fields:
            self.values[field.id] = self.get_value(field, quick)

        return self

    def __getitem__(self, item):
        return self.values[item]

    def get_value(self, field, quick):
        # type: (Field, bool) -> Any
        """ Ask user the question represented by this instance.

        Args:
            field (Field):
                The field we're asking the user to provide the value for.
            quick (bool):
                Enable quick mode. In quick mode, the form will reduce the
                number of question asked by using defaults wherever possible.
                This can greatly reduce the number of interactions required on
                the user part, but will obviously limit the user choices. This
                should probably be enabled only by a specific user action
                (like passing a ``--quick`` flag etc.).

        Returns:
            The user response converted to a python type using the
            :py:attr:`cliform.core.Field.type` converter.
        """
        if callable(field.default):
            default = field.default(self)
        else:
            default = field.default

        if quick and default is not None:
            return default

        shell.cprint('<90>{}', field.help)

        while True:
            try:
                answer = click.prompt(field.pretty_prompt, default=default)
                return field.type(answer)
            except ValueError:
                shell.cprint("<31>Unsupported value")


# Used in docstrings only until we drop python2 support
del Any, Dict, Tuple



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.fs
    :synopsis: File system related helpers.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import fnmatch
import itertools
import os
import re
from os.path import isdir, join, normpath
from typing import Iterator, List, Optional, Text, Union

# 3rd party imports
from six import string_types

# local imports
from . import conf
from . import log
from . import types
from . import context


def wrap_paths(paths):
    # type: (List[str]) -> str
    """ Put quotes around all paths and join them with space in-between. """
    if isinstance(paths, string_types):
        raise ValueError(
            "paths cannot be a string. "
            "Use array with one element instead."
        )
    return ' '.join('"' + path + '"' for path in paths)


def filtered_walk(path, include=None, exclude=None):
    # type: (str, Optional[List[str]], Optional[List[str]]) -> Iterator[str]
    """ Walk recursively starting at *path* excluding files matching *exclude*

    Args:
        path (str):
            A starting path. This has to be an existing directory.
        include (list[str]):
            A white list of glob patterns. If given, only files that match those
            globs will be yielded (filtered by exclude).
        exclude (list[str]):
            A list of glob string patterns to test against. If the file/path
            matches any of those patters, it will be filtered out.

    Returns:
        Iterator[str]: A generator yielding all the files that do not match any
        pattern in ``exclude``.
    """
    exclude = exclude or []

    if not isdir(path):
        raise ValueError("Cannot walk files, only directories")

    files = os.listdir(path)
    for name in files:
        filename = normpath(join(path, name))

        # If excluded, completely skip it. Will not recurse into directories
        if search_globs(filename, exclude):
            continue

        # If we have a whitelist and the pattern matches, yield it. If the
        # pattern didn't match and it's a dir, it will still be recursively
        # processed.
        if not include or match_globs(filename, include):
            yield filename

        if isdir(filename):
            for p in filtered_walk(filename, include, exclude):
                yield p


def match_globs(path, patterns):
    # type: (str, List[str]) -> bool
    """ Test whether the given *path* matches any patterns in *patterns*

    Args:
        path (str):
            A file path to test for matches.
        patterns (list[str]):
            A list of glob string patterns to test against. If *path* matches
            any of those patters, it will return True.

    Returns:
        bool: **True** if the *path* matches any pattern in *patterns*.
    """
    for pattern in (p for p in patterns if p):
        if pattern.startswith('/'):
            regex = fnmatch.translate(pattern[1:])

            temp_path = path[1:] if path.startswith('/') else path

            m = re.search(regex, temp_path)

            if m and m.start() == 0:
                return True

        elif fnmatch.fnmatch(path, pattern):
                return True

    return False


def search_globs(path, patterns):
    # type: (str, List[str]) -> bool
    """ Test whether the given *path* contains any patterns in *patterns*

    Args:
        path (str):
            A file path to test for matches.
        patterns (list[str]):
            A list of glob string patterns to test against. If *path* matches
            any of those patters, it will return True.

    Returns:
        bool: **True** if the ``path`` matches any pattern in *patterns*.
    """
    for pattern in (p for p in patterns if p):
        if pattern.startswith('/'):
            # If pattern starts with root it means it match from root only
            regex = fnmatch.translate(pattern[1:])
            regex = regex.replace('\\Z', '')

            temp_path = path[1:] if path.startswith('/') else path
            m = re.search(regex, temp_path)

            if m and m.start() == 0:
                return True

        else:
            regex = fnmatch.translate(pattern)
            regex = regex.replace('\\Z', '')

            if re.search(regex, path):
                return True

    return False


def write_file(path, content, mode='w'):
    # type: (Text, Union[Text,bytes], Text) -> None
    """ --pretend aware file writing.

    You can always write files manually but you should always handle the
    --pretend case.

    Args:
        path (str):
        content (str):
        mode (str):
    """
    from peltak.core import context
    from peltak.core import log

    if context.get('pretend', False):
        log.info("Would overwrite <34>{path}<32> with:\n<90>{content}",
                 path=path,
                 content=content)
    else:
        with open(path, mode) as fp:
            fp.write(content)


def collect_files(files):
    # type: (types.FilesCollection) -> List[str]
    """ Collect files using the given configuration. """
    paths = [conf.proj_path(p) for p in files.paths]

    if context.RunContext().get('verbose', 0) >= 3:
        log.info("only_staged: <33>{}".format(files.only_staged))
        log.info("untracked: <33>{}".format(files.untracked))
        log.info("whitelist: <33>\n{}".format('\n'.join(files.whitelist())))
        log.info("blacklist: <33>\n{}".format('\n'.join(files.blacklist())))

    return list(itertools.chain.from_iterable(
        filtered_walk(path, files.whitelist(), files.blacklist())
        for path in paths
    ))


# Used in type hint comments only (until we drop python2 support)
del Iterator, List, Optional, Text, Union, types



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" AppEngine commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import itertools
import sys
from fnmatch import fnmatch
from os.path import exists, join, normpath
from typing import Optional

# 3rd party imports
import attr

# local imports
from peltak.core import conf
from peltak.core import context
from peltak.core import fs
from peltak.core import git
from peltak.core import log
from peltak.core import shell
from peltak.core import versioning
from peltak.core import util


@util.mark_experimental
def deploy(app_id, version, promote, quiet):
    # type: (str, str, bool, bool) -> None
    """ Deploy the app to AppEngine.

    Args:
        app_id (str):
            AppEngine App ID. Overrides config value app_id if given.
        version (str):
            AppEngine project version. Overrides config values if given.
        promote (bool):
            If set to **True** promote the current remote app version to the one
            that's being deployed.
        quiet (bool):
            If set to **True** this will pass the ``--quiet`` flag to gcloud
            command.
    """
    gae_app = GaeApp.for_branch(git.current_branch().name)

    if gae_app is None and None in (app_id,  version):
        msg = (
            "Can't find an AppEngine app setup for branch <35>{}<32> and"
            "--project and --version were not given."
        )
        log.err(msg, git.current_branch().name)
        sys.exit(1)

    if version is not None:
        gae_app.version = version

    if app_id is not None:
        gae_app.app_id = app_id

    gae_app.deploy(promote, quiet)


@util.mark_experimental
def devserver(port, admin_port, clear):
    # type: (int, int, bool) -> None
    """ Run devserver.

    Args:
        port (int):
            Port on which the app will be served.
        admin_port (int):
            Port on which the admin interface is served.
        clear (bool):
            If set to **True**, clear the datastore on startup.
    """
    admin_port = admin_port or (port + 1)

    args = [
        '--port={}'.format(port),
        '--admin_port={}'.format(admin_port)
    ]

    if clear:
        args += ['--clear_datastore=yes']

    with conf.within_proj_dir():
        shell.run('dev_appserver.py . {args}'.format(args=' '.join(args)))


@util.mark_experimental
def setup_ci():
    # type: () -> None
    """ Setup AppEngine SDK on CircleCI """
    gcloud_path = shell.run('which gcloud', capture=True).stdout.strip()
    sdk_path = normpath(join(gcloud_path, '../../platform/google_appengine'))
    gcloud_cmd = gcloud_path + ' --quiet'

    if not exists(sdk_path):
        log.info("Installing AppEngine SDK")
        shell.run('sudo {} components install app-engine-python'.format(
            gcloud_cmd
        ))
    else:
        # Only initialise once. To reinitialise, just build without cache.
        log.info("AppEngine SDK already initialised")

    log.info("Using service account authentication")
    shell.run('{} auth activate-service-account --key-file {}'.format(
        gcloud_cmd,
        conf.proj_path('ops/client_secret.json')
    ))

    # shell.run('{} config set project {}'.format(gcloud_cmd, project))


def _find_appengine_sdk():
    # type: () -> str
    gcloud_path = shell.run('which gcloud', capture=True).stdout.strip()
    return normpath(join(gcloud_path, '../../platform/google_appengine'))


@attr.s
class GaeApp(object):
    """ Represents an AppEngine app.

    Attributes:
        app_id (str):
            AppEngine project ID.
        version (str):
            AppEngine project version.
        deployables (list[str]):
            List of YAML files to deploy. Defaults to all yaml files in the
            project root directory.
    """
    app_id = attr.ib(type=str)
    version = attr.ib(type=str, default=None)
    deployables = attr.ib(type=list, default=['.'])

    @classmethod
    def for_branch(cls, branch_name):
        # type: (str) -> Optional[GaeApp]
        """ Return app configuration for the given branch.

        This will look for the configuration in the ``appengine.projects``
        config variable.

        Args:
            branch_name (str):
                The name of the branch we want the configuration for.

        Returns:
            Optional[GaeApp]: The `GaeApp` instance with the configuration for
                the project.
            None: If no project configuration can be found.
        """
        for proj in conf.get('appengine.projects', []):
            if fnmatch(branch_name, proj['branch']):
                proj = dict(proj)
                proj.pop('branch')
                proj['deployables'] = list(frozenset(itertools.chain(
                    conf.get('appengine.deployables', []),
                    proj.get('deployables', [])
                )))
                return cls(**proj)

        return None

    @property
    def app_version(self):
        # type: () -> str
        """ AppEngine compatible app version.

        This assumes git-flow branching strategy and has a few predefined ways
        to automatically generate the version string.
        """
        branch = git.current_branch()

        if self.version is not None:
            return self.version

        elif branch.type == 'develop':
            return '{ver}-c{commit_nr}-{commit_id}'.format(
                ver=versioning.current().replace('.', '-'),
                commit_nr=git.latest_commit().number,
                commit_id=git.latest_commit().id
            )

        elif branch.type in ('feature', 'hotfix', 'task'):
            # We don't need version when we have the title.
            return branch.title.replace('_', '-')

        return versioning.current().replace('.', '-')

    def deploy(self, promote=False, quiet=False):
        # type: (bool, bool, bool) -> None
        """ Deploy the code to AppEngine.

        Args:
            promote (bool):
                Migrate the traffic to the deployed version.
            quiet (bool):
                Pass ``--quiet`` flag to gcloud command
        """
        args = [
            '--promote' if promote else '--no-promote',
            '--version {}'.format(self.app_version),
            '--project {}'.format(self.app_id),
        ]

        if quiet:
            args += ['--quiet']

        cmd = 'gcloud app deploy {args} {deployables}'.format(
            deployables=fs.wrap_paths(self.deployables),
            args=' '.join(args)
        )

        if context.get('pretend', False):
            log.info("Would deploy version <35>{ver}<32> to <35>{app}".format(
                ver=self.app_version,
                app=self.app_id
            ))
            shell.cprint('<90>{}', cmd)
        else:
            log.info("Deploying version <35>{ver}<32> to <35>{app}".format(
                ver=self.app_version,
                app=self.app_id,
            ))
            shell.run(cmd)


# Used in type hint comments only (until we drop python2 support)
del Optional



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Commands related to Google AppEngine.

Only useful on appengine projects.
"""
from __future__ import absolute_import

from peltak.commands import root_cli, click, pretend_option


@root_cli.group('appengine')
def appengine_cli():
    # type: () -> None
    """ Google AppEngine related commands. """
    pass


@appengine_cli.command('deploy')
@click.option(
    '--project',
    type=str,
    metavar='APP_ID',
    help=("The AppEngine project to deploy to. This overrides the value "
          "configured using appengine.projects config variable.")
)
@click.option(
    '--version',
    type=str,
    metavar='APP_VERSION',
    help=("The AppEngine version to deploy. This overrides the value "
          "configured using appengine.projects config variable.")
)
@click.option(
    '--promote',
    is_flag=True,
    help=("If specified, the currently deployed version will become the active "
          "one")
)
@click.option(
    '-q', '--quiet',
    is_flag=True,
    help="Do not prompt for input"
)
@pretend_option
def deploy(project, version, promote, quiet):
    """ Deploy the app to the target environment.

    The target environments can be configured using the ENVIRONMENTS conf
    variable. This will also collect all static files and compile translation
    messages
    """
    from . import logic

    logic.deploy(project, version, promote, quiet)


@appengine_cli.command()
@click.option('-p', '--port', type=int, default=8080)
@click.option('--admin-port', type=int, default=None)
@click.option('--clear', is_flag=True)
@pretend_option
def devserver(port, admin_port, clear):
    # type: (int, int, bool) -> None
    """ Run devserver. """
    from . import logic

    logic.devserver(port, admin_port, clear)


@appengine_cli.command('setup-ci')
@pretend_option
def setup_ci():
    # type: () -> None
    """ Setup AppEngine SDK on CircleCI """
    from . import logic

    logic.setup_ci()



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import re
import textwrap
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Pattern

# local imports
from peltak.core import conf
from peltak.core import git
from peltak.core import shell
from peltak.core import util
from peltak.core import versioning
from .types import ChangelogItems, ChangelogTag


@util.mark_experimental
def changelog():
    # type: () -> str
    """ Print change log since last release. """
    # Skip 'v' prefix
    versions = [x for x in git.tags() if versioning.is_valid(x[1:])]

    cmd = 'git log --format=%H'
    if versions:
        cmd += ' {}..HEAD'.format(versions[-1])

    hashes = shell.run(cmd, capture=True).stdout.strip().splitlines()
    commits = [git.CommitDetails.get(h) for h in hashes]

    tags = [
        ChangelogTag(**x) for x in
        conf.get('changelog.tags', (
            {'header': 'Features', 'tag': 'feature'},
            {'header': 'Changes', 'tag': 'change'},
            {'header': 'Fixes', 'tag': 'fix'},
        ))
    ]

    results = OrderedDict((tag.header, []) for tag in tags)    # type: ChangelogItems

    for commit in commits:
        commit_items = extract_changelog_items(commit.desc, tags)
        for header, items in commit_items.items():
            results[header] += items

    lines = [
        '<35>v{}<0>'.format(versioning.current()),
        '',
    ]
    for header, items in results.items():
        if items:
            lines += [
                '',
                '<32>{}<0>'.format(header),
                '<32>{}<0>'.format('-' * len(header)),
                '',
            ]
            for item_text in items:
                item_lines = textwrap.wrap(item_text, 77)
                lines += ['- {}'.format('\n  '.join(item_lines))]

            lines += ['']

    return '\n'.join(lines)


def extract_changelog_items(text, tags):
    # type: (str, List[ChangelogTag]) -> Dict[str, List[str]]
    """ Extract all tagged items from text.

    Args:
        text (str):
            Text to extract the tagged items from. Each tagged item is a
            paragraph that starts with a tag. It can also be a text list item.

    Returns:
        tuple[list[str], list[str], list[str]]:
            A tuple of ``(features, changes, fixes)`` extracted from the given
            text.

    The tagged items are usually features/changes/fixes but it can be configured
    through `pelconf.yaml`.
    """

    patterns = {tag.header: tag_re(tag.tag) for tag in tags}
    items = {tag.header: [] for tag in tags}    # type: ChangelogItems
    curr_tag = None
    curr_text = ''

    for line in text.splitlines():
        if not line.strip():
            if curr_tag is not None:
                items[curr_tag].append(curr_text)
                curr_text = ''
            curr_tag = None

        for tag in tags:
            m = patterns[tag.header].match(line)
            if m:
                if curr_tag is not None:
                    items[curr_tag].append(curr_text)
                    curr_text = ''

                curr_tag = tag.header
                line = m.group('text')
                break

        if curr_tag is not None:
            curr_text = '{} {}'.format(curr_text.strip(), line.strip()).strip()

    if curr_tag is not None:
        items[curr_tag].append(curr_text)

    return items


def tag_re(tag):
    # type: (str) -> Pattern
    """ Return a regular expression to match tagged lines.

    Args:
        tag (str):
            A tag for which you need the regex pattern. This should be a single
            word.

    Returns:
        Pattern: Regex patter object as returned by `re.compile`.

    The returned pattern will match only the starting line that contains the
    tag. The item definition might span across multiple lines and this has to
    be handled separately
    """
    return re.compile(
        # r'\(feature\) (?P<text>.*?\n\n)',
        r'(- |\* |\s+)?\({tag}\) (?P<text>.*)'.format(tag=tag),
    )


# Used in type hint comments only (until we drop python2 support)
del Any, Dict, List, Optional, Pattern, ChangelogItems



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" scripts logic. """

# stdlib imports
import subprocess
import sys
from typing import Any, Dict, List

# 3rd party imports
import attr
import yaml

# local imports
from peltak.commands import click
from peltak.core import conf
from peltak.core import log
from peltak.core import fs
from peltak.core import shell
from peltak.core.context import RunContext
from peltak.core import templates
from .types import CliOptions, Script


def run_script(script, options):
    # type: (Script, CliOptions) -> None
    """ Run the script with the given (command line) options. """
    template_ctx = build_template_context(script, options)
    verbose = RunContext().get('verbose')
    pretend = RunContext().get('pretend')

    if verbose >= 3:
        log.info('Compiling script <35>{name}\n{script}'.format(
            name=script.name,
            script=shell.highlight(script.command, 'jinja')
        ))
        yaml_str = yaml.dump(template_ctx, default_flow_style=False)
        log.info('with context:\n{}\n'.format(shell.highlight(yaml_str, 'yaml')))

    cmd = templates.Engine().render(script.command, template_ctx)
    retcode = exec_script_command(cmd, pretend)

    if verbose:
        log.info("Script exited with code: <33>{}", retcode)

    if retcode not in script.success_exit_codes:
        sys.exit(retcode)


def exec_script_command(cmd, pretend):
    # type: (str, bool) -> int
    """ This will execute the already compiled script command.

    This function serves the purpose of encapsulating the low level code of
    spawning a subprocess from the rest of the logic.
    """
    if not pretend:
        with conf.within_proj_dir():
            # God knows why, if we run the command using `shell.run()` and it
            # exists with non-zero code it will also kill the parent peltak
            # process. The Popen arguments are the same, to my knowledge
            # everything is the same but the way it behaves is completely
            # different. If we just use Popen directly, everything works as
            # expected  ¯\_(ツ)_/¯
            p = subprocess.Popen(cmd, shell=True)
            try:
                p.communicate()
                return p.returncode
            except KeyboardInterrupt:
                p.kill()
                return -1
    else:
        log.info(
            "<90>{bar}<0>\n{script}\n<90>{bar}",
            bar='=' * 80,
            script=shell.highlight(cmd, 'bash'),
        )
        return 0


def build_template_context(script, options):
    """ Build command template context.

    This will collect all the values like current configuration, command line
    options, runtime context etc. and pass it to the script template.
    """
    template_ctx = {
        'opts': dict(
            verbose=RunContext().get('verbose'),
            pretend=RunContext().get('pretend'),
            **options
        ),
        'script': attr.asdict(script),
        'conf': conf.values,
        'ctx': RunContext().values,
        'proj_path': conf.proj_path,
    }

    if script.files:
        template_ctx['files'] = fs.collect_files(script.files)

    return template_ctx


# Used only in type hint comments.
del Dict, Any, List, click, CliOptions, Script



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.shell
    :synopsis: Shell related helpers.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import re
import subprocess
import sys
from typing import Any, Dict

# 3rd party imports
import attr

# local imports
from . import context


EnvDict = Dict[str, str]


@attr.s
class ExecResult(object):
    """ Encapsulates a `shell.run` result.

    Attributes:
        command (str):
            The command that was executed.
        return_code (int):
            The command exit code.
        stdout (str):
            The command standard output content after the execution.
        stderr (str):
            The command standard error content after the execution.
        succeeded (bool):
            **True** if command was successful (return_code was 0 or the one
            that was expected).
        failed (bool):
            **True** if command failed.

    """
    command = attr.ib(type=str)
    return_code = attr.ib(type=int)
    stdout = attr.ib(type=str)
    stderr = attr.ib(type=str)
    succeeded = attr.ib(type=bool)
    failed = attr.ib(type=bool)


is_tty = sys.stdout.isatty()


def decolorize(text):
    # type: (str) -> str
    """ Remove all opcodes from the text.

    Args:
        text (str):
            Source text that includes color opcodes as used in `shell.fmt`
            function.

    Returns:
        str: The same text but with all opcodes removed.
    """
    return re.sub(r'<(\d{1,2})>', '', text)


def fmt(msg, *args, **kw):
    # type: (str, *Any, **Any) -> str
    """ Generate shell color opcodes from a pretty coloring syntax. """
    global is_tty

    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    opcode_subst = '\x1b[\\1m' if is_tty else ''
    return re.sub(r'<(\d{1,2})>', opcode_subst, msg)


def cprint(msg, *args, **kw):
    # type: (str, *Any, **Any) -> None
    """ Print colored message to stdout. """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    print(fmt('{}<0>'.format(msg)))


def run(cmd,
        capture=False,
        shell=True,
        env=None,
        exit_on_error=None,
        never_pretend=False):
    # type: (str, bool, bool, EnvDict, bool, bool) -> ExecResult
    """ Run a shell command.

    Args:
        cmd (str):
            The shell command to execute.
        shell (bool):
            Same as in `subprocess.Popen`.
        capture (bool):
            If set to True, it will capture the standard input/error instead of
            just piping it to the caller stdout/stderr.
        env (dict[str, str]):
            The subprocess environment variables.
        exit_on_error (bool):
            If set to **True**, on failure it will call `sys.exit` with the
            return code for the executed command.
        never_pretend (bool):
            If set to **True** the command will always be executed, even if
            context.get('pretend') is set to True. If set to **False** or not
            given, if the ``pretend`` context value is **True**, this function
            will only print the command it would execute and then return
            a fake result.

    Returns:
        ExecResult: The execution result containing the return code and output
        (if capture was set to *True*).
    """
    if context.get('pretend', False) and not never_pretend:
        cprint('<90>{}', cmd)
        return ExecResult(
            cmd,
            0,              # retcode
            '',             # stdout
            '',             # stderr
            True,           # succeeded
            False,          # failed
        )

    if context.get('verbose', 0) > 2:
        cprint('<90>{}', cmd)

    options = {
        'bufsize': 1,       # line buffered
        'shell': shell
    }   # type: Dict[str, Any]

    if exit_on_error is None:
        exit_on_error = not capture

    if capture:
        options.update({
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
        })

    if env is not None:
        options['env'] = dict(os.environ)
        options['env'].update(env)

    p = subprocess.Popen(cmd, **options)

    try:
        stdout, stderr = p.communicate()

        try:
            if stdout is not None:
                stdout = stdout.decode('utf-8')

            if stderr is not None:
                stderr = stderr.decode('utf-8')
        except AttributeError:
            # 'str' has no attribute 'decode'
            pass

        if exit_on_error and p.returncode != 0:
            sys.exit(p.returncode)

        return ExecResult(
            cmd,
            p.returncode,
            stdout,
            stderr,
            p.returncode == 0,
            p.returncode != 0
        )

    except KeyboardInterrupt:
        p.kill()
        raise


def highlight(code, fmt):
    # type: (str, str) -> str
    """ Highlight a given code snippet for printing in the terminal.

    Assumes 256 color terminal.
    """
    import pygments
    from pygments.lexers.data import YamlLexer
    from pygments.lexers.data import JsonLexer
    from pygments.lexers.python import PythonLexer
    from pygments.lexers.python import Python3Lexer
    from pygments.lexers.templates import DjangoLexer
    from pygments.lexers.shell import BashLexer
    from pygments.formatters.terminal256 import Terminal256Formatter

    # Get lexer class based on format.
    lexer_cls = {
        'yaml': YamlLexer,
        'json': JsonLexer,
        'bash': BashLexer,
        'py': Python3Lexer,
        'py3': Python3Lexer,
        'py2': PythonLexer,
        'jinja': DjangoLexer,
        'django': DjangoLexer,
    }.get(fmt)

    if not lexer_cls:
        raise ValueError("Unsupported code format: {}".format(fmt))

    return pygments.highlight(code, lexer_cls(), Terminal256Formatter())


# Used in docstrings only until we drop python2 support
del Any, Dict



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Root level CLI commands. """
from __future__ import absolute_import

# stdlib imports
from typing import Any, List

# local imports
from . import root_cli, click, pretend_option, verbose_option


@root_cli.command('clean')
@click.option(
    '-e', '--exclude',
    multiple=True,
    metavar='PATTERN',
    help='Comma separated list of paths to exclude from deletion'
)
@pretend_option
def clean(exclude):
    # type: (List[str]) -> None
    """ Remove temporary files like python cache, swap files, etc.

    You can configure the list of patterns with clean_patterns config variable.
    By default it will remove all files/dirs matching

    Config example::

        \b
        clean:
          patterns:
            - '*__pycache__*',
            - '*.py[cod]',
            - '*.swp'
          exclude:
            - '.tox'
            - '.venv'

    Examples::

        \b
        $ peltak clean
        $ peltak clean -e "*.tox*"
        $ peltak clean --pretend

    """
    from peltak.logic import root
    root.clean(exclude)


@root_cli.command('init')
@click.option(
    '-q', '--quick',
    is_flag=True,
    help="Enable quick mode. Defaults will be used wherever possible."
)
@click.option(
    '-b', '--blank',
    is_flag=True,
    help="Create a blank configuration without asking any questions."
)
@click.option(
    '-f', '--force',
    is_flag=True,
    help="Force creating. This will not ask whether to wipe out existing config"
         "or cancel in case the config already exists."
)
@pretend_option
@verbose_option
def init(**args):
    # type: (**Any) -> None
    """ Create new peltak config file in the current directory.

    If ``pelconf.py`` already exists the user will be prompted to confirm
    before continuing.

    Example::

        $ peltak init

    """
    from peltak.logic import root
    root.init(**args)


# Used in docstrings only until we drop python2 support
del Any, List



# -*- coding: utf-8 -*-
""" A place to store all patch functions specific to peltak.

Easier to track them down, when you need one.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from functools import wraps
from os.path import join
from typing import Any, Dict

# 3rd party imports
from mock import patch, Mock
from peltak.core.shell import ExecResult

# local imports
from peltak.core import shell


def patch_is_tty(value):
    """ Wrapped test function will have peltak.core.shell.is_tty set to *value*. """
    def decorator(fn):  # pylint: disable=missing-docstring
        @wraps(fn)
        def wrapper(*args, **kw):   # pylint: disable=missing-docstring
            is_tty = shell.is_tty
            shell.is_tty = value

            try:
                return fn(*args, **kw)
            finally:
                shell.is_tty = is_tty
        return wrapper

    return decorator


class patch_proj_root(object):
    """ Patch project root decorator. """
    def __init__(self, proj_root, nest_level=0):
        self.proj_root = proj_root
        self.nest = nest_level

    def __call__(self, fn):
        if self.proj_root is not None:
            cwd = join(self.proj_root, *(['fake_dir'] * self.nest))
            dirs = ['not_pelconf'] * self.nest + ['pelconf.py']
            dirs = [[d] for d in dirs]
        else:
            cwd = join('/', *(['fake_dir'] * self.nest))
            dirs = [[d] for d in ['not_pelconf'] * self.nest]

        @patch('os.getcwd', Mock(return_value=cwd))
        @patch('os.listdir', Mock(side_effect=dirs))
        @wraps(fn)
        def wrapper(*args, **kw):       # pylint: disable=missing-docstring
            return fn(*args, **kw)

        return wrapper


def patch_pelconf(config):
    # type: (Dict[str, Any]) -> Any
    """ Patch the peltak configuration.

    This will patch all content retrieved through `peltak.core.conf.get()` and
    `conf.get_path()`.

    Args:
        config (dict[str, Any]):
            The dictionary with the peltak configuration.
    """
    return patch('peltak.core.conf.values', config)


def patch_run(stdout=None, retcode=None, stderr=None, cmd=None):
    # type: (str, int, str, str) -> Any
    """ Patch shell.run and make it return a given result.

    Args:
        stdout (str):
        retcode (int):
        stderr (str):
        cmd (str):

    Returns:
        FunctionType:

    """
    p_run = Mock(return_value=ExecResult(
        cmd or '',
        retcode or 0,
        stdout or '',
        stderr or '',
        retcode == 0,
        retcode != 0,
    ))

    return patch('peltak.core.shell.run', p_run)


# Used only in type hint comments
del Any, Dict



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Types and classes used by `peltak.extra.changelog`. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any, Dict, List


# 3rd party imports
import attr


ChangelogItems = Dict[str, List[Any]]


@attr.s
class ChangelogTag(object):
    """ Changelog tag config.

    This is used to map the tags used in commit message details to the actual
    headers used in the generated changelog.

    Attributes:
        header (str):   The header used in the generated changelog.
        tag (str):      The tag used in commit message details.

    """
    header = attr.ib(type=str)
    tag = attr.ib(type=str)


# Used only in type hint comments
del Any, Dict, List



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Helper function to ease implementation of all git flow commands. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import re
import sys
from typing import List, Optional, Iterable

# 3rd party imports
import click

# local imports
from peltak.core import conf
from peltak.core import context
from peltak.core import git
from peltak.core import log
from peltak.core import shell


RE_INVALID_CHARS = re.compile('[\-\[\]\s(),.]+')


def assert_branch_type(branch_type):
    # type: (str) -> None
    """ Print error and exit if the current branch is not of a given type.

    Args:
        branch_type (str):
            The branch type. This assumes the branch is in the '<type>/<title>`
            format.
    """
    branch = git.current_branch(refresh=True)

    if branch.type != branch_type:
        if context.get('pretend', False):
            log.info("Would assert that you're on a <33>{}/*<32> branch",
                     branch_type)
        else:
            log.err("Not on a <33>{}<31> branch!", branch_type)
            fmt = ("The branch must follow <33>{required_type}/<name><31>"
                   "format and your branch is called <33>{name}<31>.")
            log.err(fmt, required_type=branch_type, name=branch.name)
            sys.exit(1)


def assert_on_branch(branch_name):
    # type: (str) -> None
    """ Print error and exit if *branch_name* is not the current branch.

    Args:
        branch_name (str):
            The supposed name of the current branch.
    """
    branch = git.current_branch(refresh=True)

    if branch.name != branch_name:
        if context.get('pretend', False):
            log.info("Would assert that you're on a <33>{}<32> branch",
                     branch_name)
        else:
            log.err("You're not on a <33>{}<31> branch!", branch_name)
            sys.exit(1)


def git_branch_delete(branch_name):
    # type: (str) -> None
    """ Delete the given branch.

    Args:
        branch_name (str):
            Name of the branch to delete.
    """
    if branch_name not in git.protected_branches():
        log.info("Deleting branch <33>{}", branch_name)
        shell.run('git branch -d {}'.format(branch_name))


def git_branch_rename(new_name):
    # type: (str) -> None
    """ Rename the current branch

    Args:
        new_name (str):
            New name for the current branch.
    """
    curr_name = git.current_branch(refresh=True).name

    if curr_name not in git.protected_branches():
        log.info("Renaming branch from <33>{}<32> to <33>{}".format(
            curr_name, new_name
        ))
        shell.run('git branch -m {}'.format(new_name))


def git_checkout(branch_name, create=False):
    # type: (str, bool) -> None
    """ Checkout or create a given branch

    Args:
        branch_name (str):
            The name of the branch to checkout or create.
        create (bool):
            If set to **True** it will create the branch instead of checking it
            out.
    """
    log.info("Checking out <33>{}".format(branch_name))
    shell.run('git checkout {} {}'.format('-b' if create else '', branch_name))


def git_pull(branch_name):
    # type: (str) -> None
    """ Pull from remote branch.

    Args:
        branch_name (str):
            The remote branch to pull.
    """
    log.info("Pulling latest changes on <33>{}", branch_name)
    shell.run('git pull origin {}'.format(branch_name))


def git_merge(base, head, no_ff=False):
    # type: (str, str, bool) -> None
    """ Merge *head* into *base*.

    Args:
        base (str):
            The base branch. *head* will be merged into this branch.
        head (str):
            The branch that will be merged into *base*.
        no_ff (bool):
            If set to **True** it will force git to create merge commit. If set
            to **False** (default) it will do a fast-forward merge if possible.
    """
    pretend = context.get('pretend', False)
    branch = git.current_branch(refresh=True)

    if branch.name != base and not pretend:
        git_checkout(base)

    args = []

    if no_ff:
        args.append('--no-ff')

    log.info("Merging <33>{}<32> into <33>{}<32>", head, base)
    shell.run('git merge {args} {branch}'.format(
        args=' '.join(args),
        branch=head,
    ))

    if branch.name != base and not pretend:
        git_checkout(branch.name)


def git_prune():
    # type: () -> None
    """ Prune dead branches. """
    log.info("Pruning")
    shell.run('git fetch --prune origin')


def get_base_branch():
    # type: () -> str
    """ Return the base branch for the current branch.

    This function will first try to guess the base branch and if it can't it
    will let the user choose the branch from the list of all local branches.

    Returns:
        str: The name of the branch the current branch is based on.
    """
    base_branch = git.guess_base_branch()

    if base_branch is None:
        log.info("Can't guess the base branch, you have to pick one yourself:")
        base_branch = choose_branch()

    return base_branch


def choose_branch(exclude=None):
    # type: (Optional[Iterable[str]]) -> str
    """ Show the user a menu to pick a branch from the existing ones.

    Args:
        exclude (list[str]):
            List of branch names to exclude from the menu. By default it will
            exclude master and develop branches. To show all branches pass an
            empty array here.

    Returns:
        str: The name of the branch chosen by the user. If the user inputs an
        invalid choice, he will be asked again (and again) until he picks a
        a valid branch.
    """
    if exclude is None:
        master = conf.get('git.master_branch', 'master')
        develop = conf.get('git.devel_branch', 'develop')
        exclude = {master, develop}

    branches = list(set(git.branches()) - set(exclude))

    # Print the menu
    for i, branch_name in enumerate(branches):
        shell.cprint('<90>[{}] <33>{}'.format(i + 1, branch_name))

    # Get a valid choice from the user
    choice = 0
    while choice < 1 or choice > len(branches):
        prompt = "Pick a base branch from the above [1-{}]".format(
            len(branches)
        )
        choice = click.prompt(prompt, value_proc=int)   # type: ignore
        if not (1 <= choice <= len(branches)):
            fmt = "Invalid choice {}, you must pick a number between {} and {}"
            log.err(fmt.format(choice, 1, len(branches)))

    return branches[choice - 1]


def to_branch_name(name):
    # type: (str) -> str
    """ Convert a given name into a valid branch name.

    This is helpful to sanitize user input before creating a new branch.

    Args:
        name (str):
            The name of the branch as provided by the user.

    Returns:
        str: The name mangled where all special characters and spaces are
        replaced by underscores.
    """
    return RE_INVALID_CHARS.sub(' ', name).strip().replace(' ', '_').lower()


# Used in docstrings only until we drop python2 support
del Iterable, List, Optional



# -*- coding: utf-8 -*-
"""
.. module:: peltak.core.pelconf
    :synopsis: `pelconf.yaml` handling.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import os.path
import sys
from contextlib import contextmanager
from types import ModuleType
from typing import Any, Dict, Iterator, List, Optional, Text, Union

# local imports
from . import hooks
from . import log
from . import util


class Pelconf(util.Singleton):
    """ Represents the `pelconf.yaml` file. """
    def __init__(self):
        if not self._singleton_initialized:
            self.values = {}    # type: Dict[str, Any]

    def from_file(self, path):
        """ Load config values from a YAML file. """
        with self.within_proj_dir():
            with open(path) as fp:
                self.values = util.yaml_load(fp) or {}

    def reset(self, config):
        """ Reset config to the given values.

        This will completely overwrite the current configuration.
        """
        self.values = config

    @classmethod
    def init(cls):
        ""
        cfg = cls()
        cfg.values = {}
        if os.path.exists('pelconf.yaml'):
            cfg.from_file('pelconf.yaml')

        # Add src_dir to sys.paths if it's set. This is only done with YAML
        # configs, py configs have to do this manually.
        src_dir = cfg.get_path('src_dir', None)
        if src_dir is not None:
            sys.path.insert(0, src_dir)

        for cmd in cfg.get('commands', []):
            try:
                _import(cmd)
            except ImportError as ex:
                log.err("Failed to load commands from <33>{}<31>: {}", cmd, ex)

        hooks.register.call('post-conf-load')

    def get(self, name, *default):
        # type: (str, *Any) -> Any
        """ Get config value with the given name and optional default.

        Args:
            name (str):
                The name of the config value.
            *default (Any):
                If given and the key doesn't not exist, this will be returned
                instead. If it's not given and the config value does not exist,
                AttributeError will be raised

        Returns:
            The requested config value. This is one of the global values defined
            in this file. If the value does not exist it will return *default* if
            give or raise `AttributeError`.

        Raises:
            AttributeError: If the value does not exist and *default* was not given.
        """
        curr = self.values
        for part in name.split('.'):
            if part in curr:
                curr = curr[part]
            elif default:
                return default[0]
            else:
                raise AttributeError("Config value '{}' does not exist".format(
                    name
                ))

        return curr

    def get_path(self, name, *default):
        # type: (str, Any) -> Any
        """ Get config value as path relative to the project directory.

        This allows easily defining the project configuration within the fabfile
        as always relative to that fabfile.

        Args:
            name (str):
                The name of the config value containing the path.
            *default (Any):
                If given and the key doesn't not exist, this will be returned
                instead. If it's not given and the config value does not exist,
                AttributeError will be raised

        Returns:
            The requested config value. This is one of the global values defined
            in this file. If the value does not exist it will return *default* if
            give or raise `AttributeError`.

        Raises:
            AttributeError: If the value does not exist and *default* was not given.
        """
        value = self.get(name, *default)

        if value is None:
            return None

        return self.proj_path(value)

    def get_env(self, name, *default):
        # type: (str, Any) -> Union[str, Any]
        """ Get the value of an ENV variable. """
        return os.environ.get(name, *default)

    def proj_path(self, *path_parts):
        # type: (*str) -> str
        """ Return absolute path to the repo dir (root project directory).

        Args:
            path (str):
                The path relative to the project root (pelconf.yaml).

        Returns:
            str: The given path converted to an absolute path.
        """
        parts = list(path_parts) or ['.']

        # If path represented by path_parts is absolute, do not modify it.
        if not os.path.isabs(parts[0]):
            proj_path = _find_proj_root()

            if proj_path is not None:
                parts = [proj_path] + list(parts)

        return os.path.normpath(os.path.join(*parts))

    @contextmanager
    def within_proj_dir(self, path='.'):
        # type: (str) -> Iterator[None]
        """ Return an absolute path to the given project relative path.

        :param path:
            Project relative path that will be converted to the system wide absolute
            path.
        :return:
            Absolute path.
        """
        curr_dir = os.getcwd()

        os.chdir(self.proj_path(path))

        yield

        os.chdir(curr_dir)


@util.cached_result()
def _find_proj_root():
    # type: () -> Optional[str]
    """ Find the project path by going up the file tree.

    This will look in the current directory and upwards for the pelconf file
    (.yaml or .py)
    """
    proj_files = frozenset(('pelconf.py', 'pelconf.yaml'))
    curr = os.getcwd()

    while curr.startswith('/') and len(curr) > 1:
        if proj_files & frozenset(os.listdir(curr)):
            return curr
        else:
            curr = os.path.dirname(curr)

    return None


def _import(cmd):
    # type: (str) -> ModuleType
    """ Exists only so we can patch it in tests."""
    return __import__(cmd)   # nocov


# Used in type hint comments only (until we drop python2 support)
del Any, Dict, Iterator, List, Optional, Union, Text, ModuleType



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
########################################
Auto-generate changelog from git history
########################################

Overview
========

This package adds the commands that allow you to greatly reduce the amount of
work required in maintaining a change log for your releases. The main command
**peltak changelog** will parse all commits since the last release and extract
all features/changes/fixes implemented. To achieve this, it will look for
specific tags in the commit descriptions. Once all those are found it will print
a nice change log. Here's an example::

    $ peltak changelog
    v0.21.2

    Features:
    - A description of a feature, taken from the commit message
    - Each feature has it's own list item
    - And each commit can have multiple features and/or fixes defined

    Changes:
    - One thing had to be change to something else.
    - Another thing also had to be changed.

    Fixes:
    - Description of the first fix
    - Each fix has it's own list item


Commit Format
=============

.. code-block:: text

    <title>

    <description text>

    (<tag>) <feature description>

    <description text>

Example
-------

.. code-block:: text

    Just added some new cool feature

    (feature) This feature allows the user to do many
    great things.
    (fix) This feature also fixes some issues we had
    before

    Closes #1


Configuration
=============

The command can be configured through `pelconf.yaml`. You can define what tags
are searched and what is their corresponding header in the changelog. Here is
an example (this is the default configuration):

.. code-block:: yaml

    changelog:
      tags:
        - header: 'Features'
          tag: 'feature'
        - header: 'Changes'
          tag: 'change'
        - header: 'Fixes'
          tag: 'fix'

"""
from __future__ import absolute_import, unicode_literals
from .commands import changelog_cli



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.git
    :synopsis: Git helpers.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
from collections import namedtuple
from typing import Any, Dict, List, Optional

# 3rd party imports
import attr
from six import string_types

# local imports
from . import conf
from . import shell
from . import util


Author = namedtuple('Author', 'name email')


@attr.s
class BranchDetails(object):
    """ Branch name parsed into type and title.

    Helpful for things like implementing git flow etc.

    Attributes:
        name (str):
            The full name of the branch as returned by git.
        type (str):
            The type of the branch. This assumes the ``type/title`` branch
            naming format and will be the text before ``/`` in the branch name.
        title (str):
            The title of the branch. This assumes the ``type/title`` branch
            naming format and will be the text after ``/`` in the branch name.
    """
    name = attr.ib(type=str)
    type = attr.ib(type=Optional[str], default=None)
    title = attr.ib(type=str, default=None)

    @classmethod
    def parse(cls, branch_name):
        # type: (str) -> BranchDetails
        """ Parse branch name into BranchDetails instance.

        Args:
            branch_name (str):
                The name of the branch to parse.

        Returns:
            BranchDetails: The parsed branch name - easy to use.
        """
        if '/' in branch_name:
            branch_type, branch_title = branch_name.rsplit('/', 1)
            return BranchDetails(branch_name, branch_type, branch_title)
        else:
            return BranchDetails(branch_name, None, branch_name)


class CommitDetails(object):
    """ Allows querying git commits for details.

    Attributes:
        id (str):
            The commit ID (first 7 characters taken from `sha1`).
        sha1 (str):
            Full SHA1 of the commit.
        author (Author):
            The commit author.
        title (str):
            The title of the commit. This is the first line of the commit
            message and should usually be shorter than 50 characters.
        desc (str):
            The commit description. This is the portion following the title.
            Will be an empty string if the commit contains only the title.
        parents_sha1 (List[str]):
            List of SHA1s of parents of this commit.
    """
    def __init__(self, sha1, author, title, desc, parents_sha1):
        self.id = sha1[:7]
        self.sha1 = sha1
        self.author = author
        self.title = title
        self.desc = desc
        self.parents_sha1 = parents_sha1
        self._branches = None
        self._parents = None

    @property
    def branches(self):
        # type: () -> List[str]
        """ List of all branches this commit is a part of. """
        if self._branches is None:
            cmd = 'git branch --contains {}'.format(self.sha1)
            out = shell.run(
                cmd,
                capture=True,
                never_pretend=True
            ).stdout.strip()
            self._branches = [x.strip('* \t\n') for x in out.splitlines()]

        return self._branches

    @property
    def parents(self):
        # type: () -> List[CommitDetails]
        """ Parents of the this commit. """
        if self._parents is None:
            self._parents = [CommitDetails.get(x) for x in self.parents_sha1]

        return self._parents

    @property
    def number(self):
        # type: () -> int
        """ Return this commits number.

        This is the same as the total number of commits in history up until
        this commit.

        This value can be useful in some CI scenarios as it allows to track
        progress on any given branch (although there can be two commits with the
        same number existing on different branches).

        Returns:
            int: The commit number/index.
        """
        cmd = 'git log --oneline {}'.format(self.sha1)
        out = shell.run(cmd, capture=True, never_pretend=True).stdout.strip()
        return len(out.splitlines())

    @classmethod
    def get(cls, sha1=''):
        # type: (str) -> CommitDetails
        """ Return details about a given commit.

        Args:
            sha1 (str):
                The sha1 of the commit to query. If not given, it will return
                the details for the latest commit.

        Returns:
            CommitDetails: Commit details. You can use the instance of the
            class to query git tree further.
        """
        with conf.within_proj_dir():
            cmd = 'git show -s --format="%H||%an||%ae||%s||%b||%P" {}'.format(
                sha1
            )
            result = shell.run(cmd, capture=True, never_pretend=True).stdout

        parts = result.split('||')
        if len(parts) != 6:
            raise ValueError('Invalid git show result:\n  {}'.format(result))

        sha1, name, email, title, desc, parents = result.split('||')

        return CommitDetails(
            sha1=sha1,
            author=Author(name, email),
            title=title,
            desc=desc,
            parents_sha1=parents.split(),
        )


@util.cached_result()
def current_branch():
    # type: () -> BranchDetails
    """ Return the BranchDetails for the current branch.

    Return:
        BranchDetails: The details of the current branch.
    """
    cmd = 'git symbolic-ref --short HEAD'
    branch_name = shell.run(
        cmd,
        capture=True,
        never_pretend=True
    ).stdout.strip()

    return BranchDetails.parse(branch_name)


@util.cached_result()
def latest_commit():
    # type: () -> CommitDetails
    """ Return details for the latest commit.

    Returns:
        CommitDetails: The `CommitDetails` instance for the latest commit on the
        current branch.
    """
    return CommitDetails.get()


def commit_details(sha1=''):
    # type: (str) -> CommitDetails
    """ Return details about a given commit.

    Args:
        sha1 (str):
            The sha1 of the commit to query. If not given, it will return the
            details for the latest commit.

    Returns:
        CommitDetails: A named tuple ``(name, email)`` with the commit author
        details.
    """
    return CommitDetails.get(sha1)


def commit_branches(sha1):
    # type: (str) -> List[str]
    """ Get the name of the branches that this commit belongs to. """
    cmd = 'git branch --contains {}'.format(sha1)
    return shell.run(
        cmd,
        capture=True,
        never_pretend=True
    ).stdout.strip().split()


@util.cached_result()
def guess_base_branch():
    # type: () -> Optional[str]
    """ Try to guess the base branch for the current branch.

    Do not trust this guess. git makes it pretty much impossible to guess
    the base branch reliably so this function implements few heuristics that
    will work on most common use cases but anything a bit crazy will probably
    trip this function.

    Returns:
        Optional[str]: The name of the base branch for the current branch if
        guessable or **None** if can't guess.
    """
    my_branch = current_branch(refresh=True).name

    curr = latest_commit()
    if len(curr.branches) > 1:
        # We're possibly at the beginning of the new branch (currently both
        # on base and new branch).
        other = [x for x in curr.branches if x != my_branch]
        if len(other) == 1:
            return other[0]
        return None
    else:
        # We're on one branch
        parent = curr

        while parent and my_branch in parent.branches:
            curr = parent

            if len(curr.branches) > 1:
                other = [x for x in curr.branches if x != my_branch]
                if len(other) == 1:
                    return other[0]
                return None

            parents = [p for p in curr.parents if my_branch in p.branches]
            num_parents = len(parents)

            if num_parents > 2:
                # More than two parent, give up
                return None
            if num_parents == 2:
                # This is a merge commit.
                for p in parents:
                    if p.branches == [my_branch]:
                        parent = p
                        break
            elif num_parents == 1:
                parent = parents[0]
            elif num_parents == 0:
                parent = None

        return None


@util.mark_deprecated('CommitDetails.get().author')
def commit_author(sha1=''):
    # type: (str) -> Author
    """ Return the author of the given commit.

    Args:
        sha1 (str):
            The sha1 of the commit to query. If not given, it will return the
            sha1 for the current commit.
    Returns:
        Author: A named tuple ``(name, email)`` with the commit author details.
    """
    with conf.within_proj_dir():
        cmd = 'git show -s --format="%an||%ae" {}'.format(sha1)
        result = shell.run(
            cmd,
            capture=True,
            never_pretend=True
        ).stdout
        name, email = result.split('||')
        return Author(name, email)


@util.cached_result()
def untracked():
    # type: () -> List[str]
    """ Return a list of untracked files in the project repository.

    Returns:
        list[str]: The list of files not tracked by project git repo.
    """
    with conf.within_proj_dir():
        status = shell.run(
            'git status --porcelain',
            capture=True,
            never_pretend=True
        ).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status.startswith('?? '):
                results.append(file_status[3:].strip())

        return results


@util.cached_result()
def unstaged():
    # type: () -> List[str]
    """ Return a list of unstaged files in the project repository.

    Returns:
        list[str]: The list of files not tracked by project git repo.
    """
    with conf.within_proj_dir():
        status = shell.run(
            'git status --porcelain',
            capture=True,
            never_pretend=True
        ).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status.strip() and file_status[0] == ' ':
                results.append(file_status[3:].strip())

        return results


@util.cached_result()
def staged():
    # type: () -> List[str]
    """ Return a list of project files staged for commit.

    Returns:
        list[str]: The list of project files staged for commit.
    """
    with conf.within_proj_dir():
        status = shell.run(
            'git status --porcelain',
            capture=True,
            never_pretend=True
        ).stdout
        results = []

        for file_status in status.split(os.linesep):
            if file_status and file_status[0] in ('A', 'M', 'D'):
                results.append(file_status[3:].strip())

        return results


@util.cached_result()
def ignore():
    # type: () -> List[str]
    """ Return a list of patterns in the project .gitignore

    Returns:
        list[str]: List of patterns set to be ignored by git.
    """

    def parse_line(line):   # pylint: disable=missing-docstring
        # Decode if necessary
        if not isinstance(line, string_types):
            line = line.decode('utf-8')

        # Strip comment
        line = line.split('#', 1)[0].strip()

        return line

    ignore_files = [
        conf.proj_path('.gitignore'),
        conf.proj_path('.git/info/exclude'),
        config().get('core.excludesfile')
    ]

    result = []     # type: List[str]
    for ignore_file in ignore_files:
        if not (ignore_file and os.path.exists(ignore_file)):
            continue

        with open(ignore_file) as fp:
            parsed = (parse_line(l) for l in fp.readlines())
            result += [x for x in parsed if x]

    return result


@util.cached_result()
def branches():
    # type: () -> List[str]
    """ Return a list of branches in the current repo.

    Returns:
        list[str]: A list of branches in the current repo.
    """
    out = shell.run(
        'git branch',
        capture=True,
        never_pretend=True
    ).stdout.strip()
    return [x.strip('* \t\n') for x in out.splitlines()]


def tag(name, message, author=None):
    # type: (str, str, Optional[Author]) -> None
    """ Tag the current commit.

    Args:
        name (str):
            The tag name.
        message (str):
            The tag message. Same as ``-m`` parameter in ``git tag``.
        author (Author):
            The commit author. Will default to the author of the commit.
        pretend (bool):
            If set to **True** it will print the full ``git tag`` command
            instead of actually executing it.
    """
    cmd = (
        'git -c "user.name={author.name}" -c "user.email={author.email}" '
        'tag -a "{name}" -m "{message}"'
    ).format(
        author=author or latest_commit().author,
        name=name,
        message=message.replace('"', '\\"').replace('`', '\\`'),
    )
    shell.run(cmd)


@util.cached_result()
def config():
    # type: () -> Dict[str, Any]
    """ Return the current git configuration.

    Returns:
        dict[str, Any]: The current git config taken from ``git config --list``.
    """
    out = shell.run(
        'git config --list',
        capture=True,
        never_pretend=True
    ).stdout.strip()

    result = {}
    for line in out.splitlines():
        name, value = line.split('=', 1)
        result[name.strip()] = value.strip()

    return result


@util.cached_result()
def tags():
    # type: () -> List[str]
    """ Returns all tags in the repo.

    Returns:
        list[str]: List of all tags in the repo, sorted as versions.

    All tags returned by this function will be parsed as if the contained
    versions (using ``v:refname`` sorting).
    """
    return shell.run(
        'git tag --sort=v:refname',
        capture=True,
        never_pretend=True
    ).stdout.strip().splitlines()


def verify_branch(branch_name):
    # type: (str) -> bool
    """ Verify if the given branch exists.

    Args:
        branch_name (str):
            The name of the branch to check.

    Returns:
        bool: **True** if a branch with name *branch_name* exits, **False**
        otherwise.
    """
    try:
        shell.run(
            'git rev-parse --verify {}'.format(branch_name),
            never_pretend=True
        )
        return True
    except IOError:
        return False


@util.cached_result()
def protected_branches():
    # type: () -> List[str]
    """ Return branches protected by deletion.

    By default those are master and devel branches as configured in pelconf.

    Returns:
        list[str]: Names of important branches that should not be deleted.
    """
    master = conf.get('git.master_branch', 'master')
    develop = conf.get('git.devel_branch', 'develop')
    return conf.get('git.protected_branches', (master, develop))


# Used in docstrings only until we drop python2 support
del Any, Dict, List, Optional



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
.. module:: peltak.core.util
    :synopsis: Various helpers that do not depend on anything else in the project.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import re
import time
import warnings
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Text,
    TextIO,
    Union
)

# 3rd party imports
import yaml
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper     # type: ignore


TextOrStream = Union[Text, TextIO]
YamlData = Union[Dict[Text, Any], List[Any]]
AnyFunction = Callable[..., Any]
Decorator = Callable[[AnyFunction], AnyFunction]


class timed_block(object):  # noqa
    """ Context manager to measure execution time for a give block of code.

    Example:
        >>> import time
        >>> from peltak.core import util
        >>>
        >>> with util.timed_block() as t:
        ...     time.sleep(1)
        >>>
        >>> print("Code executed in {}s".format(int(t.elapsed_s)))
        Code executed in 1s

    Attributes:
        t0 (float):
            The time at the start of execution.
        elapsed (float);
            Raw elapsed time in seconds.
        elapsed_s (float):
            Elapsed time in seconds, rounded to 3 decimal places for easy
            display.
        elapsed_ms (float):
            Elapsed time in milliseconds, rounded to 3 decimal places for easy
            display.
    """
    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.elapsed = time.time() - self.t0
        self.elapsed_s = round(self.elapsed, 3)
        self.elapsed_ms = round(self.elapsed * 1000, 3)


def mark_experimental(fn):
    # type: (AnyFunction) -> AnyFunction
    """ Mark function as experimental.

    Args:
        fn (FunctionType):
            The command function to decorate.
    """
    @wraps(fn)
    def wrapper(*args, **kw):   # pylint: disable=missing-docstring
        from peltak.core import shell

        if shell.is_tty:
            warnings.warn("This command is has experimental status. The "
                          "interface is not yet stable and might change "
                          "without notice. Use at your own risk")
        return fn(*args, **kw)

    return wrapper


def mark_deprecated(replaced_by):
    # type: (Text) -> Decorator
    """ Mark command as deprecated.

    Args:
        replaced_by (str):
            The command that deprecated this command and should be used instead.
    """
    def decorator(fn):   # pylint: disable=missing-docstring
        @wraps(fn)
        def wrapper(*args, **kw):   # pylint: disable=missing-docstring
            from peltak.core import shell

            if shell.is_tty:
                warnings.warn("This command is has been deprecated. Please use "
                              "'{new}' instead.".format(new=replaced_by))

            return fn(*args, **kw)

        return wrapper

    return decorator


class cached_result(object):
    """ Decorator that caches the function result.

    This is especially useful for functions whose output won't change during
    a single *peltak* execution run and are expensive (i.e. call external
    shell commands).

    Example:
        >>> from peltak.core import util
        >>>
        >>> @util.cached_result()
        ... def foo():
        ...     call_count = getattr(foo, 'call_count', 0)
        ...     call_count += 1
        ...     setattr(foo, 'call_count', call_count)
        >>> foo()
        >>> print(foo.call_count)
        1
        >>> foo()
        >>> print(foo.call_count)
        1
        >>> util.cached_result.clear(foo)
        >>> foo()
        >>> print(foo.call_count)
        2
    """
    CACHE_VAR = '__cached_result'

    def __init__(self):
        pass

    def __call__(self, fn):
        # type: (AnyFunction) -> AnyFunction
        """ Apply the decorator to the given function.

        Args:
            fn (FunctionType):
                The function to decorate.
        :return Function:
            The function wrapped in caching logic.
        """
        @wraps(fn)
        def wrapper(refresh=False):   # pylint: disable=missing-docstring
            if refresh or not hasattr(wrapper, self.CACHE_VAR):
                result = fn()
                setattr(wrapper, self.CACHE_VAR, result)

            return getattr(wrapper, self.CACHE_VAR)

        return wrapper

    @classmethod
    def clear(cls, fn):
        # type: (AnyFunction) -> None
        """ Clear result cache on the given function.

        If the function has no cached result, this call will do nothing.

        Args:
            fn (FunctionType):
                The function whose cache should be cleared.
        """
        if hasattr(fn, cls.CACHE_VAR):
            delattr(fn, cls.CACHE_VAR)


def in_batches(iterable, batch_size):
    # type: (Iterable[Any], int) -> Iterator[List[Any]]
    """ Split the given iterable into batches.

    Args:
        iterable (Iterable[Any]):
            The iterable you want to split into batches.
        batch_size (int):
            The size of each bach. The last batch will be probably smaller (if
            the number of elements cannot be equally divided.

    Returns:
        Iterator[list[Any]]: Will yield all items in batches of **batch_size**
            size.

    Example:

        >>> from peltak.core import util
        >>>
        >>> batches = util.in_batches([1, 2, 3, 4, 5, 6, 7], 3)
        >>> batches = list(batches)     # so we can query for length
        >>> len(batches)
        3
        >>> batches
        [[1, 2, 3], [4, 5, 6], [7]]

    """
    items = list(iterable)
    size = len(items)

    for i in range(0, size, batch_size):
        yield items[i:min(i + batch_size, size)]


def yaml_load(str_or_fp):
    # type: (TextOrStream) -> YamlData
    """ Load data from YAML string or file-like object.

    Args:
        str_or_fp (Union[str, TextIO]):

    Returns:
        YamlData: The data loaded from the YAML string/file.
    """
    return yaml.load(str_or_fp, Loader=Loader)


def yaml_dump(data, stream=None):
    # type: (YamlData, Optional[Any]) -> Text
    """ Dump data to a YAML string/file.

    Args:
        data (YamlData):
            The data to serialize as YAML.
        stream (TextIO):
            The file-like object to save to. If given, this function will write
            the resulting YAML to that stream.

    Returns:
        str: The YAML string.
    """
    return yaml.dump(
        data,
        stream=stream,
        Dumper=Dumper,
        default_flow_style=False
    )


def remove_indent(text):
    # type: (Text) -> Text
    """ Remove indentation from the text.

    All indentation will be removed no matter if it's consistent across the
    *text*.

    Args:
        text (str): The text that contains indentation.

    Returns:
        str: The same text but with all indentation removed from each line.
    """
    return re.sub('\n[ \t]+', '\n', text.lstrip())


class Singleton(object):
    """ Base class for singletons.

    Inheriting from this class will make your class a singleton. It should have
    an initializer that takes no arguments.
    """
    instances = {}  # type: Dict[str, Any]

    def __new__(cls, *args, **kw):
        # Could upgrade this to thread local storage
        instance = Singleton.instances.get(cls.__name__)

        if instance:
            instance._singleton_initialized = True
        else:
            instance = object.__new__(cls)
            instance._singleton_initialized = False
            Singleton.instances[cls.__name__] = instance

        return instance


# Used in type hint comments only (until we drop python2 support)
del Any, Dict, Iterable, List, Text, TextIO, Union, Iterator, Optional



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Helpers for testing peltak related code.

This is kept inside the peltak package as it ca be used by 3rd party packages
extending the app.
"""
from __future__ import absolute_import, unicode_literals


# pkg interface
from .mocks import mock_result
from .patches import patch_is_tty
from .patches import patch_proj_root
from .patches import patch_pelconf
from .patches import patch_run
from .util import TestDataProvider



# -*- coding: utf-8 -*-
""" Various utilities to help write tests for peltak. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import shutil
import tempfile
from contextlib import contextmanager
from os.path import basename, join


class TestDataProvider(object):
    """ Provides a temporary copy of the selected test project. """

    def __init__(self, data_dir):
        self.data_dir = data_dir
        self.proj_dir = None

    @contextmanager
    def __call__(self, path):
        temp_dir = tempfile.mkdtemp(prefix='peltak_tests_')

        cwd = os.getcwd()
        self.proj_dir = join(temp_dir, basename(path))
        proj_src = join(self.data_dir, path)

        shutil.copytree(proj_src,  self.proj_dir)
        os.chdir(self.proj_dir)

        try:
            yield self
        finally:
            os.chdir(cwd)
            shutil.rmtree(temp_dir)



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Commands package.

All commands (and only commands) should be defined inside this package. This
should be as thin layer as possible. Ideally just processing CLI params and
displaying results.
"""
from __future__ import absolute_import

# stdlib imports
from typing import Any, Callable, Union

# 3rd party imports
import click

# local imports
import peltak


AnyFn = Callable[..., Any]


@click.group()
@click.version_option(version=peltak.__version__, message='%(version)s')
def root_cli():
    # type: () -> None
    """

    To get help for a specific command:

       \033[1m peltak <command> --help\033[0m

    Examples:

       \033[1m peltak lint --help\033[0m

       \033[1m peltak version bump --help\033[0m

       \033[1m peltak release upload --help\033[0m

    """
    pass


def pretend_option(fn):
    # type: (AnyFn) -> AnyFn
    """ Decorator to add a --pretend option to any click command.

    The value won't be passed down to the command, but rather handled in the
    callback. The value will be accessible through `peltak.core.context` under
    'pretend' if the command needs it. To get the current value you can do:

        >>> from peltak.commands import click, root_cli
        >>> from peltak.core import context
        >>>
        >>> @root_cli.command('my-command')
        >>> @pretend_option
        >>> def my_command():
        ...     pretend = context.get('pretend', False)

    This value will be accessible from anywhere in the code.
    """

    def set_pretend(ctx, param, value):     # pylint: disable=missing-docstring
        # type: (click.Context, Union[click.Option, click.Parameter], Any) -> Any
        from peltak.core import context
        from peltak.core import shell

        context.set('pretend', value or False)
        if value:
            shell.cprint('<90>{}', _pretend_msg())

    return click.option(
        '--pretend',
        is_flag=True,
        help=("Do not actually do anything, just print shell commands that "
              "would be executed."),
        expose_value=False,
        callback=set_pretend
    )(fn)


def _pretend_msg():
    from peltak.core import util
    msg = '''
        ┌─────────────────────────────────────────────────────────────────┐
        │                     Running in pretend mode.                    │
        ├─────────────────────────────────────────────────────────────────┤
        │ All commands that you will see printed in gray are not actually │
        │ executed but printed out instead.                               │
        │                                                                 │
        │ Those would be executed if --pretend option was not specified.  │
        │                                                                 │
        │ You can still add verbosity to some commands using the          │
        │ -v/--verbose option.                                            │
        └─────────────────────────────────────────────────────────────────┘
    '''

    if hasattr(msg, 'decode'):  # py2
        msg = msg.decode('utf-8')

    try:
        # We do empty format() here so that it forces unicode errors here
        # and not when it's used/printed.
        return '{}'.format(util.remove_indent(msg))
    except:
        return util.remove_indent('''
        +-----------------------------------------------------------------+
        |                     Running in pretend mode.                    |
        +-----------------------------------------------------------------+
        | All commands that you will see printed in gray are not actually |
        | executed but printed out instead.                               |
        |                                                                 |
        | Those would be executed if --pretend option was not specified.  |
        |                                                                 |
        | You can still add verbosity to some commands using the          |
        | -v/--verbose option.                                            |
        +-----------------------------------------------------------------+
    ''')


def verbose_option(fn):
    """ Decorator to add a --verbose option to any click command.

    The value won't be passed down to the command, but rather handled in the
    callback. The value will be accessible through `peltak.core.context` under
    'verbose' if the command needs it. To get the current value you can do:

        >>> from peltak.core import context
        >>>
        >>> pretend = context.get('verbose', False)

    This value will be accessible from anywhere in the code.
    """

    def set_verbose(ctx, param, value):     # pylint: disable=missing-docstring
        # type: (click.Context, Union[click.Option, click.Parameter], Any) -> Any
        from peltak.core import context
        context.set('verbose', value or 0)

    return click.option(
        '-v', '--verbose',
        expose_value=False,
        count=True,
        callback=set_verbose,
        help="Be verbose. Can specify multiple times for more verbosity.",
    )(fn)



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Commands implementation.

Kept in a separate package to reduce the number of imports and file sizes for
the commands module. This way the commands module contain only code used by
click to generate the completion and everything else is imported when used.
"""
from __future__ import absolute_import, unicode_literals



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Git commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os

# 3rd party imports
import click

# local imports
from peltak.core import conf
from peltak.core import context
from peltak.core import fs
from peltak.core import git
from peltak.core import log
from peltak.core import shell
from peltak.core import util


def add_hooks(pre_commit, pre_push):
    # type: (str, str) -> None
    """ Add git hooks for commit and push to run linting and tests. """

    # Detect virtualenv the hooks should use

    # Detect virtualenv
    virtual_env = conf.get_env('VIRTUAL_ENV')
    if virtual_env is None:
        log.err("You are not inside a virtualenv")
        confirm_msg = (
            "Are you sure you want to use global python installation "
            "to run your git hooks? [y/N] "
        )
        click.prompt(confirm_msg, default='')
        if not click.confirm(confirm_msg):
            log.info("Cancelling")
            return

        load_venv = ''
    else:
        load_venv = 'source "{}/bin/activate"'.format(virtual_env)

    commit_hook = conf.proj_path('.git/hooks/pre-commit')
    push_hook = conf.proj_path('.git/hooks/pre-push')

    # Write pre-commit hook
    log.info("Adding pre-commit hook <33>{}", commit_hook)
    fs.write_file(commit_hook, util.remove_indent('''
        #!/bin/bash
        PATH="/opt/local/libexec/gnubin:$PATH"
        
        {load_venv}
        
        {command}
        
    '''.format(load_venv=load_venv, command=pre_commit)))

    # Write pre-push hook
    log.info("Adding pre-push hook: <33>{}", push_hook)
    fs.write_file(push_hook, util.remove_indent('''
        #!/bin/bash
        PATH="/opt/local/libexec/gnubin:$PATH"
        
        {load_venv}
        
        peltak test --allow-empty
        
        {command}
        
    '''.format(load_venv=load_venv, command=pre_push)))

    log.info("Making hooks executable")
    if not context.get('pretend', False):
        os.chmod(conf.proj_path('.git/hooks/pre-commit'), 0o755)
        os.chmod(conf.proj_path('.git/hooks/pre-push'), 0o755)


def push():
    # type: () -> None
    """ Push the current branch to origin.

    This is an equivalent of ``git push -u origin <branch>``. Mainly useful for
    the first push as afterwards ``git push`` is just quicker. Free's you from
    having to manually type the current branch name in the first push.
    """
    branch = git.current_branch().name
    shell.run('git push -u origin {}'.format(branch))



# -*- coding: utf-8 -*-
"""
.. module:: peltak.core.templates.filters
    :synopsis: Built-in script template filters

################################
Built-in script template filters
################################

Built-in jinja2 filters
=======================

You can find the list of all filters built into jinja2 in `jinja2 documentation`_

Filters provided by peltak
==========================

.. autofunction:: header
.. autofunction:: count_flag
.. autofunction:: cprint
.. autofunction:: wrap_paths


.. _jinja2 documentation:
    https://jinja.palletsprojects.com/en/2.10.x/templates/#list-of-builtin-filters

"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any

# 3rd party imports
from six import string_types

# local imports
from peltak.core import fs
from peltak.core import shell


def header(title):
    # type: (Any) -> str
    """ Converts a given title into a pretty header with colors.

    **Usage:**

    .. code-block:: jinja

        {{ 'hello' | header }}

    will result in::

        = hello ================================================================

    The resulting string will be colored for printing in the terminal.
    """
    title = str(title)
    if (len(title) > 72):
        title = title[0:70] + '...'

    remaining = 80 - len(title) - 3
    return shell.fmt('echo "<32>= <35>{title} <32>{bar}<0>"',
                     title=title,
                     bar='=' * remaining)


def count_flag(count, flag):
    # type: (int, str) -> str
    """ Returns the given flag letter as a count flag.

    **Usage:**

    .. code-block:: jinja

        {{ set verbose = 3 }}
        {{ verbose | count_flag('v') }}

    will result in::

        -vvv

    In the above example, if ``verbose`` is ``0`` the result of this filter will
    be an empty string.
    """
    if not isinstance(count, int) or count < 0:
        raise ValueError('Cannot create count flag from count={}'.format(count))

    if not isinstance(flag, string_types):
        raise ValueError("Flags must be strings, got: {}".format(flag))

    if len(flag) > 1:
        raise ValueError("Flags must have length 1, got: {}".format(flag))

    if not flag.isalpha():
        raise ValueError("Flags must be letters, got: {}".format(flag))

    return '-' + flag * count if count else ''


def cprint(msg, *args, **kw):
    """ Will convert the given message to an echo statement with color opcodes.

    This supports the same syntax as `peltak.core.shell.fmt` (used internally here).
    The color processing will replace any tag like object in format ``<NUMBER>``
    into a corresponding opcode. Here's a quick cheatsheet on some of the more
    usefull opcodes

    ========== =================================================================
     Tag        Description
    ---------- -----------------------------------------------------------------
     ``<0>``    Reset all formatting to default values.
     ``<1>``    Intensify current color (will affect all other color opcodes).
     ``<31>``   Set text color to red.
     ``<32>``   Set text color to green.
     ``<33>``   Set text color to yellow.
     ``<34>``   Set text color to blue.
     ``<35>``   Set text color to pink.
     ``<35>``   Set text color to teal.
    ========== =================================================================

    **Usage:**

    .. code-block:: jinja

        {{ '<35>hello, <32>world' | cprint }}

    will result in the following::

        echo "\\x1b[35mhello, \\x1b[32mworld\\x1b[0m"

    Which inside a terminal will be rendered as *hello* in ping and world in
    green.

    **cprint** also supports formatting, same as in the built-in ``format()``
    function.

    .. code-block:: jinja

        {{ "hello, {}, I'm {name}" | cprint('Susan', name='John') }}

    will result in::

        echo "hello, Susan, I'm John\\x1b[0m"

    """
    return shell.fmt('echo "{}<0>"', str(msg).format(*args, **kw))


def wrap_paths(paths):
    """ Returns a string with all items wrapped in quotes and separated by space.

    **Usage:**

    .. code-block:: jinja

        {% set paths = ['file 1', 'file 2', 'file 3'] %}
        cp {{ paths | wrap_paths }} out_dir/

    will result in::

        cp "file 1" "file 2" "file 3" out_dir/

    """
    return fs.wrap_paths(paths)


# Used only in type hint comments
del Any



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
##################
``peltak release``
##################
git flow release commands.
"""
from peltak.commands import root_cli, click, pretend_option


@root_cli.group('release', invoke_without_command=True)
def release_cli():
    # type: () -> None
    """ Commands that implement the release git flow.

    **Example Config**::

        \b
        version:
            file: 'src/mypkg/__init__.py'
        git:
            devel_branch: 'develop'
            master_branch: 'master'
            protected_branches: ['master', 'develop']

    Examples:

        \b
        $ peltak release start patch    # Make a new patch release
        $ peltak release start minor    # Make a new minor release
        $ peltak release start major    # Make a new major release
        $ peltak release start          # same as start patch
        $ peltak release tag            # Tag current commit as release

    """
    pass


@release_cli.command('start')
@click.argument(
    'component',
    type=click.Choice(['major', 'minor', 'patch']),
    default='patch',
)
@click.option(
    '--exact',
    type=str,
    help="Set the newly released version to be exactly as specified."
)
@pretend_option
def start(component, exact):
    # type: (str, str) -> None
    """ Create a new release.

    It will bump the current version number and create a release branch called
    `release/<version>` with one new commit (the version bump).

    **Example Config**::

        \b
        version:
            file: 'src/mypkg/__init__.py'

    **Examples**::

        \b
        $ peltak release start patch    # Make a new patch release
        $ peltak release start minor    # Make a new minor release
        $ peltak release start major    # Make a new major release
        $ peltak release start          # same as start patch

    """
    from peltak.extra.gitflow import logic
    logic.release.start(component, exact)


@release_cli.command('tag')
@click.option(
    '-m', '--message',
    type=str,
    help=("Tag message. Will replace the default 'Mark vX.X release'")
)
@pretend_option
def tag_release(message):
    # type: (str) -> None
    """ Tag the current commit with as the current version release.

    This should be the same commit as the one that's uploaded as the release
    (to pypi for example).

    **Example Config**::

        \b
        version:
            file: 'src/mypkg/__init__.py'

    Examples::

        $ peltak release tag          # Tag the current commit as release

    """
    from peltak.extra.gitflow import logic
    logic.release.tag(message)


@release_cli.command('finish')
@pretend_option
@click.option(
    '--ff', '--fast-forward', 'fast_forward',
    is_flag=True,
    help="Try to perform a fast-forward merge. If possible this will not "
         "create a merge commit on the target branch."
)
def finish(fast_forward):
    # type: (bool) -> None
    """ Merge the current release to both develop and master.

    This will perform a FF merge with develop if possible and --no-ff merge
    with master and then tag the merge commit with the current version.
    """
    from peltak.extra.gitflow import logic
    logic.release.finish(fast_forward)


@release_cli.command('merged')
@pretend_option
def merged():
    # type: () -> None
    """ Checkout the target branch, pull and delete the merged branch.

    This is to ease the repetitive cleanup of each merged branch.

    Example Config (those the are defaults)::

        \b
        git:
            devel_branch: 'develop'
            master_branch: 'master'
            protected_branches: ['master', 'develop']

    Example::

        $ peltak release merged     # Must be ran on the relase branch

    """
    from peltak.extra.gitflow import logic
    logic.release.merged()



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
##################################
**git-flow** made easy with peltak
##################################

The improvement over raw git-flow is that you will have to type way less. Most
of the time all the branch names required will be filled in.


Default configuration
====================

.. code-block:: yaml

    version_file: '/must/be/defined/in/the/project'
    git:
        devel_branch:  'develop'
        master_branch: 'master'

"""
from __future__ import absolute_import, unicode_literals

from .commands.feature import feature_cli
from .commands.hotfix import hotfix_cli
from .commands.release import release_cli



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
####################
``peltak changelog``
####################

CLI definition.
"""
from __future__ import absolute_import

from peltak.commands import root_cli, click


@root_cli.group('changelog', invoke_without_command=True)
@click.pass_context
def changelog_cli(ctx):
    # type: (click.Context) -> None
    """ Generate changelog from commit messages. """
    if ctx.invoked_subcommand:
        return

    from peltak.core import shell
    from . import logic
    shell.cprint(logic.changelog())



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" git flow hotfix commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import sys

# local imports
from peltak.core import conf
from peltak.core import context
from peltak.core import git
from peltak.core import hooks
from peltak.core import log
from peltak.core import shell
from peltak.core import versioning
from peltak.extra.changelog.logic import changelog
from . import common


def start(component, exact):
    # type: (str, str) -> None
    """ Create a new release branch.

    Args:
        component (str):
            Version component to bump when creating the release. Can be *major*,
            *minor* or *patch*.
        exact (str):
            The exact version to set for the release. Overrides the component
            argument. This allows to re-release a version if something went
            wrong with the release upload.
    """
    version_file = conf.get_path('version.file', 'VERSION')

    develop = conf.get('git.devel_branch', 'develop')
    common.assert_on_branch(develop)

    with conf.within_proj_dir():
        out = shell.run('git status --porcelain', capture=True).stdout
        lines = out.split(os.linesep)
        has_changes = any(
            not l.startswith('??') for l in lines if l.strip()
        )

    if has_changes:
        log.info("Cannot release: there are uncommitted changes")
        exit(1)

    old_ver, new_ver = versioning.bump(component, exact)

    log.info("Bumping package version")
    log.info("  old version: <35>{}".format(old_ver))
    log.info("  new version: <35>{}".format(new_ver))

    with conf.within_proj_dir():
        branch = 'release/' + new_ver

        hooks.register.call('pre-release-start', branch, old_ver, new_ver)

        common.git_checkout(branch, create=True)

        log.info("Creating commit for the release")
        shell.run('git add {ver_file} && git commit -m "{msg}"'.format(
            ver_file=version_file,
            msg="Releasing v{}".format(new_ver)
        ))

        hooks.register.call('post-release-start', branch, old_ver, new_ver)


def finish(fast_forward):
    # type: (bool) -> None
    """ Merge current release into develop and master and tag it. """
    pretend = context.get('pretend', False)

    if not pretend and (git.staged() or git.unstaged()):
        log.err(
            "You have uncommitted changes in your repo!\n"
            "You need to stash them before you merge the release branch"
        )
        sys.exit(1)

    develop = conf.get('git.devel_branch', 'develop')
    master = conf.get('git.master_branch', 'master')
    branch = git.current_branch(refresh=True)

    common.assert_branch_type('release')

    hooks.register.call('pre-release-finish', branch)

    # Merge release into master
    common.git_checkout(develop)
    common.git_pull(develop)
    common.git_merge(develop, branch.name, no_ff=not fast_forward)

    # Merge release into develop
    common.git_checkout(master)
    common.git_pull(master)
    common.git_merge(master, branch.name, no_ff=not fast_forward)

    # Tag the release commit with version number
    tag(changelog())

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(master)

    hooks.register.call('post-release-finish', branch)


def merged():
    # type: () -> None
    """ Cleanup the release branch after it was remotely merged to master. """
    develop = conf.get('git.devel_branch', 'develop')
    master = conf.get('git.master_branch', 'master')
    branch = git.current_branch(refresh=True)

    common.assert_branch_type('release')

    hooks.register.call('pre-release-merged', branch)

    # Pull master with the merged release
    common.git_checkout(master)
    common.git_pull(master)

    # Merge to develop
    common.git_checkout(develop)
    common.git_pull(develop)
    common.git_merge(develop, branch.name)

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(develop)
    hooks.register.call('post-release-merged', branch)


def tag(message):
    # type: (str) -> None
    """ Tag the current commit with the current version. """
    release_ver = versioning.current()
    message = message or 'v{} release'.format(release_ver)

    with conf.within_proj_dir():
        log.info("Creating release tag")
        git.tag(
            author=git.latest_commit().author,
            name='v{}'.format(release_ver),
            message=message,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" git flow feature commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys

# local imports
from peltak.core import conf
from peltak.core import context
from peltak.core import git
from peltak.core import hooks
from peltak.core import log
from . import common


def start(name):
    # type: (str) -> None
    """ Start working on a new feature by branching off develop.

    This will create a new branch off develop called feature/<name>.

    Args:
        name (str):
            The name of the new feature.
    """
    feature_name = 'feature/' + common.to_branch_name(name)
    develop = conf.get('git.devel_branch', 'develop')

    common.assert_on_branch(develop)

    hooks.register.call('pre-feature-start', name)
    common.git_checkout(feature_name, create=True)
    hooks.register.call('post-feature-start', name)


def update():
    # type: () -> None
    """ Update the feature with updates committed to develop.

    This will merge current develop into the current branch.
    """
    branch = git.current_branch(refresh=True)
    develop = conf.get('git.devel_branch', 'develop')

    common.assert_branch_type('feature')
    common.git_checkout(develop)
    common.git_pull(develop)
    common.git_checkout(branch.name)
    common.git_merge(branch.name, develop)


def rename(name):
    # type: (str) -> None
    """ Give the currently developed feature a new name.

    Args:
        name (str):
            The new name of the current feature. The current branch will be
            renamed to 'feature/<new_name>'.
    """
    common.assert_branch_type('feature')
    common.git_branch_rename('feature/' + name.strip().replace(' ', '_'))


def finish(fast_forward):
    # type: (bool) -> None
    """ Merge current feature branch into develop. """
    pretend = context.get('pretend', False)

    if not pretend and (git.staged() or git.unstaged()):
        log.err(
            "You have uncommitted changes in your repo!\n"
            "You need to stash them before you merge the feature branch"
        )
        sys.exit(1)

    develop = conf.get('git.devel_branch', 'develop')
    branch = git.current_branch(refresh=True)

    common.assert_branch_type('feature')

    hooks.register.call('pre-feature-finish', branch)

    # Merge feature into develop
    common.git_checkout(develop)
    common.git_pull(develop)
    common.git_merge(develop, branch.name, no_ff=not fast_forward)

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(develop)

    hooks.register.call('post-feature-finish', branch)


def merged():
    # type: () -> None
    """ Cleanup a remotely merged branch. """
    develop = conf.get('git.devel_branch', 'develop')
    branch = git.current_branch(refresh=True)

    common.assert_branch_type('feature')

    hooks.register.call('pre-feature-merged', branch)

    # Pull develop with the merged feature
    common.git_checkout(develop)
    common.git_pull(develop)

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(develop)

    hooks.register.call('post-feature-merged', branch)



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
##############
``peltak git``
##############

git helpers.
"""
from __future__ import absolute_import
from peltak.commands import click, root_cli, pretend_option


@root_cli.group('git')
def git_cli():
    # type: () -> None
    """ Git related commands """
    pass


@git_cli.command('add-hooks')
@pretend_option
@click.option(
    '--pre-commit',
    type=str,
    prompt='Pre commit tool',
    help=(
        'Command that will be ran before every commit. If it fails, the commit '
        'will fail.'
    )
)
@click.option(
    '--pre-push',
    type=str,
    prompt='Pre push tool',
    help = (
        'Command that will be ran before every push. If it fails, the push '
        'will fail.'
    )
)
def add_hooks(pre_commit, pre_push):
    # type: (str, str) -> None
    """ Setup project git hooks.

    This will run all the checks before pushing to avoid waiting for the CI
    fail.

    Example::

        $ peltak git add-hooks
    """
    from . import impl

    impl.add_hooks(pre_commit, pre_push)


@git_cli.command('push')
@pretend_option
def push():
    # type: () -> None
    """ Push the current branch and set to track remote.

    Example::

        $ peltak git push

    """
    from . import impl

    impl.push()



# -*- coding: utf-8 -*-
""" Types and classes used across **peltak** codebase. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any, Callable, Dict, List

# 3rd party imports
import attr

# local imports
from six import string_types


AnyFn = Callable[..., Any]
YamlConf = Dict[str, Any]
CliOptions = Dict[str, Any]


@attr.s
class FilesCollection(object):
    """ Configure files passed to the script.

    This section allows you to inject a set of files into the script as the
    ``{{ files }}`` expression. This configuration allows you to flexibly define
    what files are collected by specifying the *paths* of interest and further
    refining the list by *include* and *exclude* which act as a pattern based
    whitelist and blacklist for the paths specified in *paths*.

    On top of that you also have 2 boolean flags: *commit* will collect
    only files staged for commit and *untracked* (``True`` by default will
    include or not files untracked by git).
    """
    paths = attr.ib(type=List[str])
    include = attr.ib(type=List[str], factory=list)
    exclude = attr.ib(type=List[str], factory=list)
    only_staged = attr.ib(type=bool, default=False)
    untracked = attr.ib(type=bool, default=True)
    use_gitignore = attr.ib(type=bool, default=True)

    @classmethod
    def from_config(cls, files_conf):
        # type: (YamlConf) -> FilesCollection
        """ Load from config dict """
        paths = files_conf.get('paths')
        fields = attr.fields(cls)

        include = files_conf.get('include',
                                 fields.include.default.factory())  # type: ignore
        exclude = files_conf.get('exclude',
                                 fields.exclude.default.factory())  # type: ignore

        if not paths:
            raise ValueError("You must define the paths when using script files")

        # A string value is the same as one element array.
        paths = [paths] if isinstance(paths, string_types) else paths
        include = [include] if isinstance(include, string_types) else include
        exclude = [exclude] if isinstance(exclude, string_types) else exclude

        return cls(
            paths=paths,
            include=include,
            exclude=exclude,
            only_staged=files_conf.get('only_staged', fields.only_staged.default),
            untracked=files_conf.get('untracked', fields.untracked.default),
            use_gitignore=files_conf.get('use_gitignore',
                                         fields.use_gitignore.default),
        )

    def whitelist(self):
        # type: () -> List[str]
        """ Return a full whitelist for use with `fs.filtered_walk()` """
        from peltak.core import fs      # avoid circular dependency.
        from peltak.core import git

        if self.only_staged:
            # Only include committed files if commit only is true.
            include = [
                '*' + f for f in git.staged()
                if not self.include or fs.match_globs(f, self.include)
            ]
        else:
            include = list(self.include)

        return include

    def blacklist(self):
        # type: () -> List[str]
        """ Return a full blacklist for use with `fs.filtered_walk()` """
        from peltak.core import git

        exclude = list(self.exclude)

        # prepare
        if self.use_gitignore:
            exclude += git.ignore()

        if not self.untracked:
            exclude += git.untracked()

        return exclude



# -*- coding: utf-8 -*-
""" A place to store all mocks specific to peltak.

Easier to track them down, when you need one.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any, Dict, List, Union

# local imports
from peltak.core.shell import ExecResult


def mock_result(stdout=None, retcode=None, stderr=None, cmd=None):
    # type: (str, int, str, str) -> ExecResult
    """ Helper for creating ExecResults for tests.

    Args:
        stdout (str):
        retcode (int):
        stderr (str):
        cmd (str):

    Returns:
        ExecResult:
    """
    return ExecResult(
        cmd or '',
        retcode or 0,
        stdout or '',
        stderr or '',
        retcode == 0,
        retcode != 0,
    )


# Used only in type hint comments
del Any, Dict, List, Union



# -*- coding: utf-8 -*-
# Copyright 2017-2018 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Helper commands for dealing with pypi. """
from __future__ import absolute_import
from peltak.commands import click, root_cli, pretend_option


@root_cli.group('pypi')
def pypi_cli():
    # type: () -> None
    """ pypi related commands """
    pass


@pypi_cli.command('upload')
@click.argument('target', required=False, default='pypi')
@pretend_option
def upload(target):
    # type: (str) -> None
    """ Upload to a given pypi target.

    Examples::

        \b
        $ peltak pypi upload         # Upload the current code to pypi
        $ peltak pypi upload private # Upload to pypi target 'private'

    """
    from peltak.extra.pypi import logic
    logic.upload(target)


@pypi_cli.command('configure')
@click.option(
    '-u', '--username',
    type=str,
    help="PyPi username. Defaults to PYPI_USER env variable."
)
@click.option(
    '-p', '--password',
    type=str,
    help="PyPi password. Defaults to PYPI_PASS env variable."
)
@pretend_option
def configure(username, password):
    # type: (str, str) -> None
    """
    Generate .pypirc config with the given credentials.

    Example:

        $ peltak pypi configure my_pypi_user my_pypi_pass

    """
    from peltak.extra.pypi import logic
    logic.gen_pypirc(username, password)
